# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import os
import pytest
from uuid import uuid4 as random_uuid

from datacube.model import LineageDirection, LineageTree, InconsistentLineageException
from datacube.model.lineage import LineageRelations, LineageIDPair
from datacube.utils import read_documents


def test_directions():
    assert LineageDirection.SOURCES != LineageDirection.DERIVED
    assert LineageDirection.DERIVED == LineageDirection.SOURCES.opposite()
    assert LineageDirection.SOURCES == LineageDirection.DERIVED.opposite()


def test_ltree_clsmethods(data_folder):
    root = random_uuid()
    # Minimal tree - root node only
    minimal = LineageTree.from_data(dsid=root)
    assert minimal.dataset_id == root
    assert minimal.direction == LineageDirection.SOURCES
    assert minimal.children is None
    # Check optional args to from_eo3_doc are set correctly.
    optional_args = LineageTree.from_data(dsid=root, sources={},
                                          direction=LineageDirection.DERIVED,
                                          home="notused")
    assert optional_args == LineageTree(direction=LineageDirection.DERIVED,
                                        dataset_id=root,
                                        children={},
                                        home=None)   # Note home is not written to the root node
    doc = list(
        read_documents(
            str(os.path.join(data_folder, "ds_eo3.yml"))
        )
    )[0][1]
    tree = LineageTree.from_eo3_doc(doc, home="src_home", home_derived="der_home")
    assert tree.home == "der_home"
    for child in tree.children["bc"]:
        assert child.home == "src_home"


@pytest.fixture
def shared_tree_ids():
    return {
        "ard1": random_uuid(),

        "l1_1": random_uuid(),
        "l1_2": random_uuid(),
        "l1_3": random_uuid(),

        "atmos": random_uuid(),
    }


@pytest.fixture
def src_tree_ids(shared_tree_ids):
    return {
        "root": random_uuid(),
        "ard1": shared_tree_ids["ard1"],

        "l1_1": shared_tree_ids["l1_1"],
        "l1_2": shared_tree_ids["l1_2"],
        "l1_3": shared_tree_ids["l1_3"],

        "atmos": shared_tree_ids["atmos"],
    }


@pytest.fixture
def src_lineage_tree(src_tree_ids):
    ids = src_tree_ids
    direction = LineageDirection.SOURCES
    return LineageTree(
        dataset_id=ids["root"], direction=direction,
        children={
            "ard": [
                LineageTree(
                    dataset_id=ids["ard1"], direction=direction,
                    children={
                        "l1": [
                            LineageTree(
                                dataset_id=ids["l1_1"], direction=direction,
                                home="level1db",
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_2"], direction=direction,
                                home="level1db",
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_3"], direction=direction,
                                home="level1db",
                                children={}
                            ),
                        ],
                        "atmos_corr": [
                            LineageTree(
                                dataset_id=ids["atmos"], direction=direction,
                                home="level1db",
                                children=None
                            )
                        ],
                    }
                ),
            ]
        }
    )


def test_lineage_serialisation(src_lineage_tree, src_tree_ids):
    ids = src_tree_ids
    serialised = src_lineage_tree.serialise()
    assert serialised == {
        "id": str(ids["root"]),
        "sources": {
            "ard": [
                {
                    "id": str(ids["ard1"]),
                    "sources": {
                        "l1": [
                            {
                                "id": str(ids["l1_1"]),
                                "home": "level1db"
                            },
                            {
                                "id": str(ids["l1_2"]),
                                "home": "level1db"
                            },
                            {
                                "id": str(ids["l1_3"]),
                                "home": "level1db"
                            },
                        ],
                        "atmos_corr": [
                            {
                                "id": str(ids["atmos"]),
                                "home": "level1db"
                            }
                        ]
                    },
                }
            ]
        }
    }
    tree_out = LineageTree.deserialise(serialised)
    assert tree_out == src_lineage_tree


@pytest.fixture
def src_lineage_tree_diffhome(src_tree_ids):
    ids = src_tree_ids
    direction = LineageDirection.SOURCES
    return LineageTree(
        dataset_id=ids["root"], direction=direction,
        children={
            "ard": [
                LineageTree(
                    dataset_id=ids["ard1"], direction=direction,
                    children={
                        "l1": [
                            LineageTree(
                                dataset_id=ids["l1_1"], direction=direction,
                                home="elsewhere",
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_2"], direction=direction,
                                home="elsewhere",
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_3"], direction=direction,
                                children={}
                            ),
                        ],
                        "atmos_corr": [
                            LineageTree(
                                dataset_id=ids["atmos"], direction=direction,
                                children=None
                            )
                        ],
                    }
                ),
            ]
        }
    )


@pytest.fixture
def mixed_dir_lineage_tree(src_tree_ids):
    ids = src_tree_ids
    direction = LineageDirection.SOURCES
    return LineageTree(
        dataset_id=ids["root"], direction=direction,
        children={
            "ard": [
                LineageTree(
                    dataset_id=ids["ard1"], direction=LineageDirection.DERIVED,
                    children={
                        "l1": [
                            LineageTree(
                                dataset_id=ids["l1_1"], direction=direction,
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_2"], direction=direction,
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_3"], direction=direction,
                                children={}
                            ),
                        ],
                        "atmos_corr": [
                            LineageTree(
                                dataset_id=ids["atmos"], direction=direction,
                                children=None
                            )
                        ],
                    }
                ),
            ]
        }
    )


@pytest.fixture
def big_src_tree_ids(shared_tree_ids):
    ids = shared_tree_ids
    return {
        "root": random_uuid(),
        "ard1": ids["ard1"],
        "ard2": random_uuid(),

        "l1_1": ids["l1_1"],
        "l1_2": ids["l1_2"],
        "l1_3": ids["l1_3"],

        "l1_4": random_uuid(),
        "l1_5": random_uuid(),
        "l1_6": random_uuid(),

        "atmos": ids["atmos"],
        "atmos_parent": random_uuid()
    }


@pytest.fixture
def big_src_lineage_tree(big_src_tree_ids):
    ids = big_src_tree_ids
    direction = LineageDirection.SOURCES
    return LineageTree(
        dataset_id=ids["root"], direction=direction,
        children={
            "ard": [
                LineageTree(
                    dataset_id=ids["ard1"], direction=direction,
                    children={
                        "l1": [
                            LineageTree(
                                dataset_id=ids["l1_1"], direction=direction,
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_2"], direction=direction,
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_3"], direction=direction,
                                children={}
                            ),
                        ],
                        "atmos_corr": [
                            LineageTree(
                                dataset_id=ids["atmos"], direction=direction,
                                children=None
                            )
                        ],
                    }
                ),
                LineageTree(
                    dataset_id=ids["ard2"], direction=direction,
                    children={
                        "l1": [
                            LineageTree(
                                dataset_id=ids["l1_4"], direction=direction,
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_5"], direction=direction,
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_6"], direction=direction,
                                children={}
                            ),
                        ],
                        "atmos_corr": [
                            LineageTree(
                                dataset_id=ids["atmos"], direction=direction,
                                children={
                                    "preatmos": [
                                        LineageTree(
                                            dataset_id=ids["atmos_parent"], direction=direction,
                                            children={}
                                        )
                                    ]
                                }
                            )
                        ],
                    }
                ),
            ]
        }
    )


@pytest.fixture
def classifier_mismatch(big_src_tree_ids):
    ids = big_src_tree_ids
    direction = LineageDirection.SOURCES
    return LineageTree(
        dataset_id=ids["root"], direction=direction,
        children={
            "ard": [
                LineageTree(
                    dataset_id=ids["ard1"], direction=direction,
                    children={
                        "l1": [
                            LineageTree(
                                dataset_id=ids["l1_1"], direction=direction,
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_2"], direction=direction,
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_3"], direction=direction,
                                children={}
                            ),
                        ],
                        "atmos_corr": [
                            LineageTree(
                                dataset_id=ids["atmos"], direction=direction,
                                children=None
                            )
                        ],
                    }
                ),
                LineageTree(
                    dataset_id=ids["ard2"], direction=direction,
                    children={
                        "lvl1": [
                            LineageTree(
                                dataset_id=ids["l1_4"], direction=direction,
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_5"], direction=direction,
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_6"], direction=direction,
                                children={}
                            ),
                        ],
                        "atmos_corr": [
                            LineageTree(
                                dataset_id=ids["atmos"], direction=direction,
                                children={
                                    "preatmos": [
                                        LineageTree(
                                            dataset_id=ids["atmos_parent"], direction=direction,
                                            children={}
                                        )
                                    ]
                                }
                            )
                        ],
                    }
                ),
            ]
        }
    )


@pytest.fixture
def src_lineage_tree_with_bad_diamond(big_src_tree_ids):
    ids = big_src_tree_ids
    direction = LineageDirection.SOURCES
    # This is a test tree with a malformed diamond relationship.
    return LineageTree(
        dataset_id=ids["root"], direction=direction,
        children={
            "ard": [
                LineageTree(
                    dataset_id=ids["ard1"], direction=direction,
                    children={
                        "l1": [
                            LineageTree(
                                dataset_id=ids["l1_1"], direction=direction,
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_2"], direction=direction,
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_3"], direction=direction,
                                children={}
                            ),
                        ],
                        "atmos_corr": [
                            # First half of the diamond relationship:
                            # ard1 depends on atmos (which depends on atmos_parent)
                            LineageTree(
                                dataset_id=ids["atmos"], direction=direction,
                                children={
                                    "preatmos": [
                                        LineageTree(
                                            dataset_id=ids["atmos_parent"], direction=direction,
                                            children={}
                                        )
                                    ]
                                }
                            )
                        ],
                    }
                ),
                LineageTree(
                    dataset_id=ids["ard2"], direction=direction,
                    children={
                        "l1": [
                            LineageTree(
                                dataset_id=ids["l1_4"], direction=direction,
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_5"], direction=direction,
                                children={}
                            ),
                            LineageTree(
                                dataset_id=ids["l1_6"], direction=direction,
                                children={}
                            ),
                        ],
                        # Second half of the diamond relationship:
                        # ard2 also depends on atmos (which depends on atmos_parent)
                        # But the atmos->atmos_parent relationship was already recorded
                        # above, so "children" here should be None or {}.
                        "atmos_corr": [
                            LineageTree(
                                dataset_id=ids["atmos"], direction=direction,
                                children={
                                    "preatmos": [
                                        LineageTree(
                                            dataset_id=ids["atmos_parent"], direction=direction,
                                            children={}
                                        )
                                    ]
                                }
                            )
                        ],
                    }
                ),
            ]
        }
    )


def test_child_datasets(big_src_lineage_tree, big_src_tree_ids):
    cds = big_src_lineage_tree.child_datasets()
    for dsid in big_src_tree_ids.values():
        assert dsid == big_src_lineage_tree.dataset_id or dsid in cds


def test_lin_rels_lin_tree_conversions(big_src_lineage_tree, big_src_tree_ids):
    # Create LRS from LT
    rels1 = LineageRelations(tree=big_src_lineage_tree)
    # Extract LT from LRS
    extracted_tree = rels1.extract_tree(big_src_tree_ids["root"], big_src_lineage_tree.direction)
    # Confirm extract LT produces same LRS as original LT
    rels2 = LineageRelations(tree=extracted_tree)
    for rel in rels1.relations:
        assert rel in rels2.relations
    # Extract reverse direction tree
    derived_tree = rels1.extract_tree(root=big_src_tree_ids["atmos"], direction=LineageDirection.DERIVED)
    rels3 = LineageRelations(tree=derived_tree)
    extracted_tree = rels3.extract_tree(big_src_tree_ids["root"], direction=big_src_lineage_tree.direction)
    rels4 = LineageRelations(tree=extracted_tree)
    for rel in rels4.relations:
        assert rel in rels1.relations


def test_detect_cyclic_deps(big_src_lineage_tree, big_src_tree_ids):
    # Confirm trivial cyclic dependencies are detected.
    repeated_uuid = random_uuid()
    looped_tree = LineageTree(
        dataset_id=repeated_uuid, direction=LineageDirection.SOURCES,
        children={
            "cyclic_self": [
                LineageTree(dataset_id=repeated_uuid, direction=LineageDirection.SOURCES, children={})
            ]
        }
    )
    with pytest.raises(InconsistentLineageException) as e:
        rels2 = LineageRelations(tree=looped_tree)
    assert "LineageTrees must be acyclic" in str(e.value)
    # Confirm more complex cyclic dependencies can be detected.
    rels = LineageRelations(tree=big_src_lineage_tree)
    breaking_tree = LineageTree(
        dataset_id=big_src_tree_ids["l1_1"],
        direction=LineageDirection.SOURCES,
        children={
            "cyclic_dep": [LineageTree(
                dataset_id=big_src_tree_ids["root"],
                direction=LineageDirection.SOURCES
            )]
        }
    )
    with pytest.raises(InconsistentLineageException) as e:
        rels.merge_tree(breaking_tree)
    assert "LineageTrees must be acyclic" in str(e.value)


def test_subtree(big_src_lineage_tree, big_src_tree_ids, src_lineage_tree, src_tree_ids):
    sub = big_src_lineage_tree.find_subtree(big_src_tree_ids["root"])
    assert sub == big_src_lineage_tree

    # Test multiple nodes, one with children
    sub = big_src_lineage_tree.find_subtree(big_src_tree_ids["atmos"])
    assert sub.dataset_id == big_src_tree_ids["atmos"]
    assert sub.children is not None

    # Test no nodes with children
    sub = src_lineage_tree.find_subtree((src_tree_ids["atmos"]))
    assert sub.dataset_id == src_tree_ids["atmos"]
    assert sub.children is None


def test_good_consistency_check(big_src_lineage_tree, src_lineage_tree, big_src_tree_ids):
    rels1 = LineageRelations(tree=src_lineage_tree)
    rels2 = LineageRelations(tree=big_src_lineage_tree)
    diff = rels1.relations_diff(rels2)
    assert diff[1] == {} and diff[3] == {}
    assert LineageIDPair(derived_id=src_lineage_tree.dataset_id, source_id=big_src_tree_ids["ard1"]) in diff[0]
    diff = rels1.relations_diff(rels2, allow_updates=True)
    assert diff[1] == {} and diff[3] == {}
    assert LineageIDPair(derived_id=src_lineage_tree.dataset_id, source_id=big_src_tree_ids["ard1"]) in diff[0]
    diff = rels1.relations_diff()
    assert diff[1] == {} and diff[3] == {}
    assert LineageIDPair(derived_id=src_lineage_tree.dataset_id, source_id=big_src_tree_ids["ard1"]) in diff[0]


def test_bad_diamond(src_lineage_tree_with_bad_diamond, big_src_tree_ids):
    # Test detection of trees that have a diamond relationship in which both paths are extended.
    # E.g. If A->B->C->D  and A->E->C->D, the C->D relationship should only be recorded
    # under one branch (B or A) and the other occurrence of C should have no children recorded.
    with pytest.raises(InconsistentLineageException, match="Duplicate nodes in LineageTree"):
        rels = LineageRelations(tree=src_lineage_tree_with_bad_diamond)


def test_home_mismatch(big_src_lineage_tree):
    tree = big_src_lineage_tree
    tree.children["ard"][0].children["atmos_corr"][0].home = "bungalow"
    tree.children["ard"][1].children["atmos_corr"][0].home = "apartment"
    with pytest.raises(InconsistentLineageException, match="Tree contains inconsistent homes"):
        rels = LineageRelations(tree=tree)


def test_classifier_mismatch(big_src_lineage_tree, classifier_mismatch):
    rels1 = LineageRelations(tree=big_src_lineage_tree)
    rels2 = LineageRelations(tree=classifier_mismatch)
    with pytest.raises(
            InconsistentLineageException,
            match="Dataset .* is derived from .* with inconsistent classifiers."):
        rels1.merge(rels2)


def test_classifier_update(big_src_lineage_tree, classifier_mismatch):
    rels1 = LineageRelations(tree=big_src_lineage_tree)
    rels2 = LineageRelations(tree=classifier_mismatch)
    diffs = rels1.relations_diff(existing_relations=rels2, allow_updates=True)
    assert len(diffs[1]) > 0


def test_home_update(src_lineage_tree, src_lineage_tree_diffhome):
    rels1 = LineageRelations(tree=src_lineage_tree)
    rels2 = LineageRelations(tree=src_lineage_tree_diffhome)
    diffs = rels1.relations_diff(existing_relations=rels2, allow_updates=True)
    assert len(diffs[3]) > 0


def test_mixed_dirs(mixed_dir_lineage_tree):
    with pytest.raises(InconsistentLineageException, match="Tree contains both derived and source nodes"):
        rels1 = LineageRelations(tree=mixed_dir_lineage_tree)


def test_merge_tree_limited_depth(big_src_lineage_tree, big_src_tree_ids):
    ids = big_src_tree_ids
    rels = LineageRelations(tree=big_src_lineage_tree, max_depth=1)
    assert ids["root"] in rels.dataset_ids
    assert ids["ard1"] in rels.dataset_ids
    assert ids["l1_1"] not in rels.dataset_ids
    assert ids["atmos"] not in rels.dataset_ids
    assert ids["atmos_parent"] not in rels.dataset_ids

    rels = LineageRelations(tree=big_src_lineage_tree, max_depth=2)
    assert ids["root"] in rels.dataset_ids
    assert ids["ard1"] in rels.dataset_ids
    assert ids["l1_1"] in rels.dataset_ids
    assert ids["atmos"] in rels.dataset_ids
    assert ids["atmos_parent"] not in rels.dataset_ids

    rels = LineageRelations(tree=big_src_lineage_tree, max_depth=3)
    assert ids["root"] in rels.dataset_ids
    assert ids["ard1"] in rels.dataset_ids
    assert ids["l1_1"] in rels.dataset_ids
    assert ids["atmos"] in rels.dataset_ids
    assert ids["atmos_parent"] in rels.dataset_ids

    rels = LineageRelations(tree=big_src_lineage_tree, max_depth=7)
    assert ids["atmos_parent"] in rels.dataset_ids



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import datetime
import logging
from abc import ABC, abstractmethod
from time import monotonic
from typing import cast, Iterable, Sequence, Iterator, TYPE_CHECKING

from odc.geo import CRS, Geometry

from datacube.model import MetadataType, Product, QueryField, QueryDict
from datacube.utils import jsonify_document, InvalidDocException
from datacube.utils.changes import DocumentMismatchError, check_doc_unchanged, Change
from datacube.utils.documents import JsonDict, JsonLike, UnknownMetadataType

from ._types import BatchStatus

if TYPE_CHECKING:
    from ._index import AbstractIndex

_LOG = logging.getLogger(__name__)


class AbstractProductResource(ABC):
    """
    Abstract base class for the Product portion of an index api.

    All ProductResource implementations should inherit from this base
    class and implement all abstract methods.

    (If a particular abstract method is not applicable for a particular implementation
    raise a NotImplementedError)
    """
    def __init__(self, index: "AbstractIndex"):
        self._index = index

    def from_doc(self, definition: JsonDict,
                 metadata_type_cache: dict[str, MetadataType] | None = None) -> Product:
        """
        Construct unpersisted Product model from product metadata dictionary

        :param definition: a Product metadata dictionary
        :param metadata_type_cache: a dict cache of MetaDataTypes to use in constructing a Product.
                                    MetaDataTypes may come from a different index.
        :return: Unpersisted product model
        """
        # This column duplication is getting out of hand:
        Product.validate(definition)   # type: ignore[attr-defined]   # validate method added by decorator
        # Validate extra dimension metadata
        Product.validate_extra_dims(definition)

        metadata_type_in: str | JsonLike = definition['metadata_type']

        # They either specified the name of a metadata type, or specified a metadata type.
        # Is it a name?
        if isinstance(metadata_type_in, str):
            if metadata_type_cache is not None and metadata_type_in in metadata_type_cache:
                metadata_type: MetadataType | None = metadata_type_cache[metadata_type_in]
            else:
                metadata_type = self._index.metadata_types.get_by_name(metadata_type_in)
                if (metadata_type is not None
                        and metadata_type_cache is not None
                        and metadata_type.name not in metadata_type_cache):
                    metadata_type_cache[metadata_type.name] = metadata_type
        else:
            # Otherwise they embedded a document, add it if needed:
            metadata_type = self._index.metadata_types.from_doc(cast(JsonDict, metadata_type_in))
            definition = dict(definition)
            definition['metadata_type'] = metadata_type.name

        if metadata_type is None:
            raise UnknownMetadataType('Unknown metadata type: %r' % definition['metadata_type'])

        return Product(metadata_type, definition)

    @abstractmethod
    def add(self,
            product: Product,
            allow_table_lock: bool = False
           ) -> Product:
        """
        Add a product to the index.

        :param metadata_type: Unpersisted Product model
        :param allow_table_lock:
            Allow an exclusive lock to be taken on the table while creating the indexes.
            This will halt other user's requests until completed.

            If false, creation will be slightly slower and cannot be done in a transaction.

            raise NotImplementedError if set to True, and this behaviour is not applicable
            for the implementing driver.
        :return: Persisted Product model.
        """

    def _add_batch(self, batch_products: Iterable[Product]) -> BatchStatus:
        """
        Add a single "batch" of products.

        Default implementation is simple loop of add

        API Note: This API method is not finalised and may be subject to change.

        :param batch_types: An iterable of one batch's worth of Product objects to add
        :return: BatchStatus named tuple.
        """
        b_skipped = 0
        b_added = 0
        b_started = monotonic()
        for prod in batch_products:
            try:
                self.add(prod)
                b_added += 1
            except DocumentMismatchError as e:
                _LOG.warning("%s: Skipping", str(e))
                b_skipped += 1
            except Exception as e:
                _LOG.warning("%s: Skipping", str(e))
                b_skipped += 1
        return BatchStatus(b_added, b_skipped, monotonic() - b_started)

    def bulk_add(self,
                 product_docs: Iterable[JsonDict],
                 metadata_types: dict[str, MetadataType] | None = None,
                 batch_size: int = 1000) -> BatchStatus:
        """
        Add a group of product documents in bulk.

        API Note: This API method is not finalised and may be subject to change.

        :param product_docs: An iterable of product metadata docs.
        :param batch_size: Number of products to add per batch (default 1000)
        :param metadata_types: Optional dictionary cache of MetadataType objects.
                               Used for product metadata validation, and for filtering.
                               (Metadata types not in in this list are skipped.)
        :return: BatchStatus named tuple, with `safe` containing a list of
                 product names that are safe to include in a subsequent dataset bulk add.
        """
        n_in_batch = 0
        added = 0
        skipped = 0
        batch = []
        started = monotonic()
        safe = set()
        batched = set()
        existing = {prod.name: prod for prod in self.get_all()}
        for doc in product_docs:
            if metadata_types is not None:
                if doc["metadata_type"] not in metadata_types:
                    skipped += 1
                    continue
            try:
                prod = self.from_doc(doc, metadata_type_cache=metadata_types)
                if prod.name in existing:
                    check_doc_unchanged(prod.definition, jsonify_document(doc), f"Product {prod.name}")
                    _LOG.warning("%s: skipped (already loaded)", prod.name)
                    skipped += 1
                    safe.add(prod.name)
                else:
                    batch.append(prod)
                    n_in_batch += 1
                    batched.add(prod.name)
            except UnknownMetadataType as e:
                skipped += 1
            except InvalidDocException as e:
                _LOG.warning("%s: Skipped", str(e))
                skipped += 1
            if n_in_batch >= batch_size:
                batch_results = self._add_batch(batch)
                added += batch_results.completed
                skipped += batch_results.skipped
                if batch_results.safe is not None:
                    safe.update(batch_results.safe)
                else:
                    safe.update(batched)
                batched = set()
                batch = []
                n_in_batch = 0
        if n_in_batch > 0:
            batch_results = self._add_batch(batch)
            added += batch_results.completed
            skipped += batch_results.skipped
            if batch_results.safe is not None:
                safe.update(batch_results.safe)
            else:
                safe.update(batched)

        return BatchStatus(added, skipped, monotonic() - started, safe)

    @abstractmethod
    def can_update(self,
                   product: Product,
                   allow_unsafe_updates: bool = False,
                   allow_table_lock: bool = False
                  ) -> tuple[bool, Iterable[Change], Iterable[Change]]:
        """
        Check if product can be updated. Return bool,safe_changes,unsafe_changes

        (An unsafe change is anything that may potentially make the product
        incompatible with existing datasets of that type)

        :param product: product to update
        :param allow_unsafe_updates: Allow unsafe changes. Use with caution.
        :param allow_table_lock:
            Allow an exclusive lock to be taken on the table while creating the indexes.
            This will halt other user's requests until completed.

            If false, creation will be slower and cannot be done in a transaction.
        :return: Tuple of: boolean (can/can't update); safe changes; unsafe changes
        """

    @abstractmethod
    def update(self,
               product: Product,
               allow_unsafe_updates: bool = False,
               allow_table_lock: bool = False
               ) -> Product:
        """
        Persist updates to a product. Unsafe changes will throw a ValueError by default.

        (An unsafe change is anything that may potentially make the product
        incompatible with existing datasets of that type)

        :param product: Product model with unpersisted updates
        :param allow_unsafe_updates: Allow unsafe changes. Use with caution.
        :param allow_table_lock:
            Allow an exclusive lock to be taken on the table while creating the indexes.
            This will halt other user's requests until completed.

            If false, creation will be slower and cannot be done in a transaction.
        :return: Persisted updated Product model
        """

    def update_document(self,
                        definition: JsonDict,
                        allow_unsafe_updates: bool = False,
                        allow_table_lock: bool = False
                        ) -> Product:
        """
        Update a metadata type from a document. Unsafe changes will throw a ValueError by default.

        Safe updates currently allow new search fields to be added, description to be changed.

        :param definition: Updated definition
        :param allow_unsafe_updates: Allow unsafe changes. Use with caution.
        :param allow_table_lock:
            Allow an exclusive lock to be taken on the table while creating the indexes.
            This will halt other user's requests until completed.

            If false, creation will be slower and cannot be done in a transaction.
        :return: Persisted updated Product model
        """
        return self.update(self.from_doc(definition),
                           allow_unsafe_updates=allow_unsafe_updates,
                           allow_table_lock=allow_table_lock
                          )

    def add_document(self, definition: JsonDict) -> Product:
        """
        Add a Product using its definition

        :param dict definition: product definition document
        :return: Persisted Product model
        """
        type_ = self.from_doc(definition)
        return self.add(type_)

    @abstractmethod
    def delete(self, products: Iterable[Product], allow_delete_active: bool = False) -> Sequence[Product]:
        """
        Delete the specified products.

        :param products: Products to be deleted
        :param bool allow_delete_active:
            Whether to allow the deletion of a Product with active datasets
            (and thereby said active datasets). Use with caution.

            If false (default), will error if a Product has active datasets.
        :return: list of deleted Products
        """

    def get(self, id_: int) -> Product | None:
        """
        Fetch product by id.

        :param id_: Id of desired product
        :return: Product model or None if not found
        """
        try:
            return self.get_unsafe(id_)
        except KeyError:
            return None

    def get_by_name(self, name: str) -> Product | None:
        """
        Fetch product by name.

        :param name: Name of desired product
        :return: Product model or None if not found
        """
        try:
            return self.get_by_name_unsafe(name)
        except KeyError:
            return None

    @abstractmethod
    def get_unsafe(self, id_: int) -> Product:
        """
        Fetch product by id

        :param id_: id of desired product
        :return: product model
        :raises KeyError: if not found
        """

    @abstractmethod
    def get_by_name_unsafe(self, name: str) -> Product:
        """
        Fetch product by name

        :param name: name of desired product
        :return: product model
        :raises KeyError: if not found
        """

    def get_with_fields(self, field_names: Iterable[str]) -> Iterable[Product]:
        """
        Return products that have all of the given fields.

        :param field_names: names of fields that returned products must have
        :returns: Matching product models
        """
        return self.get_with_types(self._index.metadata_types.get_with_fields(field_names))

    def get_with_types(self, types: Iterable[MetadataType]) -> Iterable[Product]:
        """
        Return all products for given metadata types

        :param types: An iterable of MetadataType models
        :return: An iterable of Product models
        """
        mdts = set(mdt.name for mdt in types)
        for prod in self.get_all():
            if prod.metadata_type.name in mdts:
                yield prod

    def get_field_names(self, product: str | Product | None = None) -> Iterable[str]:
        """
        Get the list of possible search fields for a Product (or all products)

        :param product: Name of product, a Product object, or None for all products
        :return: All possible search field names
        """
        if product is None:
            prods = self.get_all()
        else:
            if isinstance(product, str):
                product = self.get_by_name(product)
            if product is None:
                prods = []
            else:
                prods = [product]
        out: set[str] = set()
        for prod in prods:
            out.update(prod.metadata_type.dataset_fields)
        return out

    def search(self, **query: QueryField) -> Iterator[Product]:
        """
        Return products that match the supplied query

        :param query: Query parameters
        :return: Generator of product models
        """
        for type_, q in self.search_robust(**query):
            if not q:
                yield type_

    @abstractmethod
    def search_robust(self,
                      **query: QueryField
                     ) -> Iterable[tuple[Product, QueryDict]]:
        """
        Return dataset types that match match-able fields and dict of remaining un-matchable fields.

        :param query: Query parameters
        :return: Tuples of product model and a dict of remaining unmatchable fields
        """

    @abstractmethod
    def search_by_metadata(self,
                           metadata: JsonDict
                           ) -> Iterable[Product]:
        """
        Perform a search using arbitrary metadata, returning results as Product objects.

        Caution – slow! This will usually not use indexes.

        :param metadata: metadata dictionary representing arbitrary search query
        :return: Matching product models
        """

    @abstractmethod
    def get_all(self) -> Iterable[Product]:
        """
        Retrieve all Products

        :returns: Product models for all known products
        """

    def get_all_docs(self) -> Iterable[JsonDict]:
        """
        Retrieve all Product metadata documents
        Default implementation calls get_all()

        API Note: This API method is not finalised and may be subject to change.

        :returns: Iterable of metadata documents for all known products
        """
        for prod in self.get_all():
            yield cast(JsonDict, prod.definition)

    @abstractmethod
    def spatial_extent(self, product: str | Product, crs: CRS = CRS("EPSG:4326")) -> Geometry | None:
        """
        Return the combined spatial extent of the nominated product

        Uses spatial index.

        Returns None if no index for the CRS, or if no datasets for the product in the relevant spatial index,
        or if the driver does not support the spatial index api.

        Result will not include extents of datasets that cannot be validly projected into the CRS.

        :param product: A Product or product name. (or None)
        :param crs: A CRS (defaults to EPSG:4326)
        :return: The combined spatial extents of the product.
        """

    @abstractmethod
    def temporal_extent(self, product: str | Product) -> tuple[datetime.datetime, datetime.datetime]:
        """
        Returns the minimum and maximum acquisition time of a product.
        Raises KeyError if product is not found, RuntimeError if product has no datasets in the index

        :param product: Product or name of product
        :return: minimum and maximum acquisition times
        """

    @abstractmethod
    def most_recent_change(self, product: str | Product) -> datetime.datetime | None:
        """
        Finds the time of the latest change to a dataset belonging to the product.
        Raises KeyError if product is not in the index
        Returns None if product has no datasets in the index

        :param product: Product or name of product
        :return: datetime of most recent dataset change
        """



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import datetime
import logging
import re
import warnings
from itertools import chain

from deprecat import deprecat
from collections import namedtuple
from time import monotonic
from typing import (Any, Callable, Iterable, Mapping, Sequence, cast)
from uuid import UUID

from datacube.migration import ODC2DeprecationWarning
from datacube.index import fields
from datacube.index.abstract import (AbstractDatasetResource, DSID, BatchStatus,
                                     NoLineageResource,
                                     dsid_to_uuid, DatasetSpatialMixin,
                                     AbstractIndex)
from datacube.model._base import QueryField
from datacube.index.fields import Field
from datacube.index.memory._fields import build_custom_fields, get_dataset_fields
from datacube.model import Dataset, LineageRelation, Product, Range, ranges_overlap
from datacube.utils import jsonify_document, _readable_offset
from datacube.utils import changes
from datacube.utils.changes import AllowPolicy, Change, Offset, get_doc_changes
from datacube.utils.dates import tz_aware
from datacube.utils.documents import metadata_subset, JsonDict

_LOG = logging.getLogger(__name__)


class DatasetResource(AbstractDatasetResource):
    def __init__(self, index: AbstractIndex) -> None:
        super().__init__(index)
        # Main dataset index
        self._by_id: dict[UUID, Dataset] = {}
        # Indexes for active and archived datasets
        self._active_by_id: dict[UUID, Dataset] = {}
        self._archived_by_id: dict[UUID, Dataset] = {}
        # Lineage indexes:
        self._derived_from: dict[UUID, dict[str, UUID]] = {}
        self._derivations: dict[UUID, dict[str, UUID]] = {}
        # Location registers
        self._locations: dict[UUID, list[str]] = {}
        self._archived_locations: dict[UUID, list[tuple[str, datetime.datetime]]] = {}
        # Active Index By Product
        self._by_product: dict[str, set[UUID]] = {}
        self._archived_by_product: dict[str, set[UUID]] = {}

    def get_unsafe(self, id_: DSID, include_sources: bool = False,
                   include_deriveds: bool = False, max_depth: int = 0) -> Dataset:
        self._check_get_legacy(include_deriveds, max_depth)
        ds = self.clone(self._by_id[dsid_to_uuid(id_)])  # N.B. raises KeyError if id not in index.
        if include_sources:
            ds.sources = {
                classifier: cast(Dataset, self.get(dsid, include_sources=True))
                for classifier, dsid in self._derived_from.get(ds.id, {}).items()
            }
        return ds

    def bulk_get(self, ids: Iterable[DSID]) -> Iterable[Dataset]:
        return (ds for ds in (self.get(dsid) for dsid in ids) if ds is not None)

    def get_derived(self, id_: DSID) -> Iterable[Dataset]:
        return (cast(Dataset, self.get(dsid)) for dsid in self._derivations.get(dsid_to_uuid(id_), {}).values())

    def has(self, id_: DSID) -> bool:
        return dsid_to_uuid(id_) in self._by_id

    def bulk_has(self, ids_: Iterable[DSID]) -> Iterable[bool]:
        return (self.has(id_) for id_ in ids_)

    def add(self, dataset: Dataset,
            with_lineage: bool = True,
            archive_less_mature: int | None = None) -> Dataset:
        if with_lineage is None:
            with_lineage = True
        _LOG.info('indexing %s', dataset.id)
        if with_lineage and dataset.sources:
            # Add base dataset without lineage
            self.add(dataset, with_lineage=False)
            # Add lineage
            for classifier, src in dataset.sources.items():
                # Recursively add source dataset and lineage
                self.add(src, with_lineage=True)
                self.persist_source_relationship(dataset, src, classifier)
        else:
            if self.has(dataset.id):
                _LOG.warning("Dataset %s is already in the database", dataset.id)
                return dataset
            persistable = self.clone(dataset, for_save=True)
            self._by_id[persistable.id] = persistable
            self._active_by_id[persistable.id] = persistable
            if dataset._uris:
                self._locations[persistable.id] = dataset._uris.copy()
            else:
                self._locations[persistable.id] = []
            self._archived_locations[persistable.id] = []
            if dataset.product.name in self._by_product:
                self._by_product[dataset.product.name].add(dataset.id)
            else:
                self._by_product[dataset.product.name] = {dataset.id}
        if archive_less_mature is not None:
            _LOG.warning("archive-less-mature functionality is not implemented for memory driver")
        return cast(Dataset, self.get(dataset.id))

    def persist_source_relationship(self, ds: Dataset, src: Dataset, classifier: str) -> None:
        # Add source lineage link
        if ds.id not in self._derived_from:
            self._derived_from[ds.id] = {}
        if self._derived_from[ds.id].get(classifier, src.id) != src.id:
            _LOG.warning("Dataset %s: Old %s dataset source %s getting overwritten by %s",
                         ds.id,
                         classifier,
                         self._derived_from[ds.id][classifier],
                         src.id)
        self._derived_from[ds.id][classifier] = src.id
        # Add source back-link
        if src.id not in self._derivations:
            self._derivations[src.id] = {}
        if self._derivations[src.id].get(classifier, ds.id) != ds.id:
            _LOG.warning("Dataset %s: Old %s dataset derivation %s getting overwritten by %s",
                         src.id,
                         classifier,
                         self._derivations[src.id][classifier],
                         ds.id)
        self._derivations[src.id][classifier] = ds.id

    def search_product_duplicates(self,
                                  product: Product,
                                  *args: str | Field
                                  ) -> Iterable[tuple[tuple, Iterable[UUID]]]:
        """
        Find dataset ids of a given product that have duplicates of the given set of field names.
        Returns each set of those field values and the datasets that have them.
        Note that this implementation does not account for slight timestamp discrepancies.
        """
        def to_field(f: str | Field) -> Field:
            if isinstance(f, str):
                f = product.metadata_type.dataset_fields[f]
            assert isinstance(f, Field), "Not a field: %r" % (f,)
            return f

        fields = [to_field(f) for f in args]
        # Typing note: mypy cannot handle dynamically created namedtuples
        GroupedVals = namedtuple('search_result', list(f.name for f in fields))  # type: ignore[misc]

        def values(ds: Dataset) -> GroupedVals:
            vals = []
            for field in fields:
                vals.append(field.extract(ds.metadata_doc))  # type: ignore[attr-defined]
            return GroupedVals(*vals)

        dups: dict[tuple, set[UUID]] = {}
        for ds in self._active_by_id.values():
            if ds.product.name != product.name:
                continue
            vals = values(ds)
            if vals in dups:
                dups[vals].add(ds.id)
            else:
                dups[vals] = set([ds.id])  # avoid duplicate entries
        # only return entries with more than one dataset
        return list({k: v for k, v in dups.items() if len(v) > 1})

    def can_update(self,
                   dataset: Dataset,
                   updates_allowed: Mapping[Offset, AllowPolicy] | None = None
                  ) -> tuple[bool, Iterable[Change], Iterable[Change]]:
        # Current exactly the same as postgres implementation.  Could be pushed up to base class?
        existing = self.get(dataset.id, include_sources=dataset.sources is not None)
        if not existing:
            raise ValueError(
                f'Unknown dataset {dataset.id}, cannot update - did you intend to add it?'
            )
        if dataset.product.name != existing.product.name:
            raise ValueError(
                'Changing product is not supported. '
                f'From {existing.product.name} to {dataset.product.name} in {dataset.id}'
            )
        # TODO: Determine (un)safe changes from metadata type
        allowed: dict[Offset, AllowPolicy] = {
            tuple(): changes.allow_extension
        }
        allowed.update(updates_allowed or {})
        doc_changes = get_doc_changes(
            existing.metadata_doc,
            jsonify_document(dataset.metadata_doc)
        )
        good_changes, bad_changes = changes.classify_changes(doc_changes, allowed)
        return not bad_changes, good_changes, bad_changes

    def update(self,
               dataset: Dataset,
               updates_allowed: Mapping[Offset, AllowPolicy] | None = None,
               archive_less_mature: int | None = None
              ) -> Dataset:
        existing = self.get(dataset.id)
        if not existing:
            raise ValueError(
                f'Unknown dataset {dataset.id}, cannot update - did you intend to add it?'
            )
        elif existing.is_archived:
            raise ValueError(f"Dataset {dataset.id} is archived.  Please restore before updating.")
        can_update, safe_changes, unsafe_changes = self.can_update(dataset, updates_allowed)
        if not safe_changes and not unsafe_changes:
            self._update_locations(dataset, existing)
            _LOG.info("No metadata changes detected for dataset %s", dataset.id)
            return dataset

        for offset, old_val, new_val in safe_changes:
            _LOG.info(
                "Safe metadata changes in %s from %r to %r",
                _readable_offset(offset),
                old_val,
                new_val
            )
        for offset, old_val, new_val in safe_changes:
            _LOG.warning(
                "Unsafe metadata changes in %s from %r to %r",
                _readable_offset(offset),
                old_val,
                new_val
            )

        if not can_update:
            unsafe_txt = ", ".join(_readable_offset(offset) for offset, _, _ in unsafe_changes)
            raise ValueError(f"Unsafe metadata changes in {dataset.id}: {unsafe_txt}")

        # Apply update
        _LOG.info("Updating dataset %s", dataset.id)
        self._update_locations(dataset, existing)
        persistable = self.clone(dataset, for_save=True)
        self._by_id[dataset.id] = persistable
        self._active_by_id[dataset.id] = persistable
        if archive_less_mature is not None:
            _LOG.warning("archive-less-mature functionality is not implemented for memory driver")
        return cast(Dataset, self.get(dataset.id))

    def _update_locations(self,
                          dataset: Dataset,
                          existing: Dataset | None = None
                         ) -> bool:
        skip_set: set[str | None] = set([None])
        new_uris: list[str] = []
        if existing and existing.uris:
            for uri in existing.uris:
                skip_set.add(uri)
            if dataset.uris:
                new_uris = [uri for uri in dataset.uris if uri not in skip_set]
        if len(new_uris):
            _LOG.info("Adding locations for dataset %s: %s", dataset.id, ", ".join(new_uris))
        for uri in reversed(new_uris):
            self.add_location(dataset.id, uri)
        return len(new_uris) > 0

    def archive(self, ids: Iterable[DSID]) -> None:
        for id_ in ids:
            id_ = dsid_to_uuid(id_)
            if id_ in self._active_by_id:
                ds = self._active_by_id.pop(id_)
                self._by_product[ds.product.name].remove(ds.id)
                if ds.product.name not in self._archived_by_product:
                    self._archived_by_product[ds.product.name] = set([ds.id])
                else:
                    self._archived_by_product[ds.product.name].add(ds.id)
                ds.archived_time = datetime.datetime.now()
                self._archived_by_id[id_] = ds

    def restore(self, ids: Iterable[DSID]) -> None:
        for id_ in ids:
            id_ = dsid_to_uuid(id_)
            if id_ in self._archived_by_id:
                ds = self._archived_by_id.pop(id_)
                ds.archived_time = None
                self._active_by_id[id_] = ds
                self._archived_by_product[ds.product.name].remove(ds.id)
                self._by_product[ds.product.name].add(ds.id)

    def purge(self, ids: Iterable[DSID], allow_delete_active: bool = False) -> Sequence[DSID]:
        purged = []
        for id_ in ids:
            id_ = dsid_to_uuid(id_)
            if id_ in self._by_id:
                ds = self._by_id.pop(id_)
                if id_ in self._active_by_id:
                    if not allow_delete_active:
                        _LOG.warning(f"Cannot purge unarchived dataset: {id_}")
                        continue
                    del self._active_by_id[id_]
                    self._by_product[ds.product.name].remove(id_)
                if id_ in self._archived_by_id:
                    del self._archived_by_id[id_]
                    self._archived_by_product[ds.product.name].remove(id_)
                if id_ in self._derived_from:
                    for classifier, src_id in self._derived_from[id_].items():
                        del self._derivations[src_id][classifier]
                    del self._derived_from[id_]
                if id_ in self._derivations:
                    for classifier, child_id in self._derivations[id_].items():
                        del self._derived_from[child_id][classifier]
                    del self._derivations[id_]
                purged.append(id_)
        return purged

    def get_all_dataset_ids(self, archived: bool | None = False) -> Iterable[UUID]:
        if archived:
            return (id_ for id_ in self._archived_by_id.keys())
        elif archived is not None:
            return (id_ for id_ in self._active_by_id.keys())
        else:
            return (id_ for id_ in self._by_id.keys())

    @deprecat(
        reason="Multiple locations per dataset are now deprecated.  Please use the 'get_location' method.",
        version="1.9.0",
        category=ODC2DeprecationWarning
    )
    def get_locations(self, id_: DSID) -> Iterable[str]:
        uuid = dsid_to_uuid(id_)
        return (s for s in self._locations[uuid])

    def get_location(self, id_: DSID) -> str | None:
        uuid = dsid_to_uuid(id_)
        locations = [s for s in self._locations.get(uuid, [])]
        if not locations:
            return None
        return locations[0]

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
               "Archived locations may not be accessible in future releases.",
        version="1.9.0",
        category=ODC2DeprecationWarning
    )
    def get_archived_locations(self, id_: DSID) -> Iterable[str]:
        uuid = dsid_to_uuid(id_)
        return (s for s, dt in self._archived_locations[uuid])

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
               "Archived locations may not be accessible in future releases.",
        version="1.9.0",
        category=ODC2DeprecationWarning
    )
    def get_archived_location_times(self, id_: DSID) -> Iterable[tuple[str, datetime.datetime]]:
        uuid = dsid_to_uuid(id_)
        return ((s, dt) for s, dt in self._archived_locations[uuid])

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
               "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning
    )
    def add_location(self, id_: DSID, uri: str) -> bool:
        uuid = dsid_to_uuid(id_)
        if uuid not in self._by_id:
            warnings.warn(f"dataset {id_} is not an active dataset")
            return False
        if not uri:
            warnings.warn(f"Cannot add empty uri. (dataset {id_})")
            return False
        if uri in self._locations[uuid]:
            return False
        self._locations[uuid].append(uri)
        return True

    def get_datasets_for_location(self, uri: str, mode: str | None = None) -> Iterable[Dataset]:
        if mode is None:
            mode = 'exact' if uri.count('#') > 0 else 'prefix'
        if mode not in ("exact", "prefix"):
            raise ValueError(f"Unsupported query mode: {mode}")
        ids: set[DSID] = set()
        if mode == "exact":
            test: Callable[[str], bool] = lambda l: l == uri  # noqa: E741
        else:
            test = lambda l: l.startswith(uri)  # noqa: E741,E731
        for id_, locs in self._locations.items():
            for loc in locs:
                if test(loc):
                    ids.add(id_)
                    break
        return self.bulk_get(ids)

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
               "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning
    )
    def remove_location(self, id_: DSID, uri: str) -> bool:
        uuid = dsid_to_uuid(id_)
        removed = False
        if uuid in self._locations:
            old_locations = self._locations[uuid]
            new_locations = [loc for loc in old_locations if loc != uri]
            if len(new_locations) != len(old_locations):
                self._locations[uuid] = new_locations
                removed = True
        if not removed and uuid in self._archived_locations:
            archived_locations = self._archived_locations[uuid]
            new_archived_locations = [(loc, dt) for loc, dt in archived_locations if loc != uri]
            if len(new_archived_locations) != len(archived_locations):
                self._archived_locations[uuid] = new_archived_locations
                removed = True
        return removed

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
               "Archived locations may not be accessible in future releases. "
               "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning
    )
    def archive_location(self, id_: DSID, uri: str) -> bool:
        uuid = dsid_to_uuid(id_)
        if uuid not in self._locations:
            return False
        old_locations = self._locations[uuid]
        new_locations = [loc for loc in old_locations if loc != uri]
        if len(new_locations) == len(old_locations):
            return False
        self._locations[uuid] = new_locations
        self._archived_locations[uuid].append((uri, datetime.datetime.now()))
        return True

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
               "Archived locations may not be restorable in future releases. "
               "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning
    )
    def restore_location(self, id_: DSID, uri: str) -> bool:
        uuid = dsid_to_uuid(id_)
        if uuid not in self._archived_locations:
            return False
        old_locations = self._archived_locations[uuid]
        new_locations = [(loc, dt) for loc, dt in old_locations if loc != uri]
        if len(new_locations) == len(old_locations):
            return False
        self._archived_locations[uuid] = new_locations
        self._locations[uuid].append(uri)
        return True

    def search_by_metadata(self, metadata: JsonDict, archived: bool | None = False) -> Iterable[Dataset]:
        if archived:
            # True: Return archived datasets only
            dss: Iterable[Dataset] = self._archived_by_id.values()
        elif archived is not None:
            # False: Return active datasets only
            dss = self._active_by_id.values()
        else:
            # True: Return archived datasets only
            dss = chain(self._active_by_id.values(), self._archived_by_id.values())
        for ds in dss:
            if metadata_subset(metadata, ds.metadata_doc):
                yield ds

    RET_FORMAT_DATASETS = 0
    RET_FORMAT_PRODUCT_GROUPED = 1

    def _search(
            self,
            return_format: int,
            limit: int | None = None,
            source_filter: Mapping[str, QueryField] | None = None,
            archived: bool | None = False,
            **query: QueryField
    ) -> Iterable[Dataset | tuple[Iterable[Dataset], Product]]:
        if "geopolygon" in query:
            raise NotImplementedError("Spatial search index API not supported by this index.")
        if source_filter:
            product_queries = list(self._get_prod_queries(**source_filter))
            if not product_queries:
                raise ValueError(f"No products match source filter: {source_filter}")
            if len(product_queries) > 1:
                raise RuntimeError("Multiproduct source_filters are not supported. Try adding 'product' field.")
            source_queries, source_product = product_queries[0]
            source_exprs = tuple(fields.to_expressions(source_product.metadata_type.dataset_fields.get,
                                                       **source_queries))
        else:
            source_product = None
            source_exprs = ()
        product_queries = list(self._get_prod_queries(**query))
        if not product_queries:
            prod_name = query.get('product')
            if prod_name is None:
                raise ValueError(f'No products match search terms: {query}')
            else:
                raise ValueError(f'No such product: {prod_name}')

        matches = 0
        for q, product in product_queries:
            if limit is not None and matches >= limit:
                break
            query_exprs = tuple(fields.to_expressions(product.metadata_type.dataset_fields.get, **q))
            product_results = []

            if archived is None:
                dsids: Iterable[UUID] = chain(
                    self._archived_by_product.get(product.name, set()),
                    self._by_product.get(product.name, set())
                )
            elif archived:
                dsids = self._archived_by_product.get(product.name, set())
            else:
                dsids = self._by_product.get(product.name, set())

            for dsid in dsids:
                if limit is not None and matches >= limit:
                    break
                ds = cast(Dataset, self.get(dsid, include_sources=True))
                query_matches = True
                for expr in query_exprs:
                    if not expr.evaluate(ds.metadata_doc):
                        query_matches = False
                        break
                if not query_matches:
                    continue
                if source_product:
                    matching_source = None
                    for sds in cast(Mapping[str, Dataset], ds.sources).values():
                        if sds.product != source_product:
                            continue
                        source_matches = True
                        for expr in source_exprs:
                            if not expr.evaluate(sds.metadata_doc):
                                source_matches = False
                                break
                        if source_matches:
                            matching_source = sds
                            break
                    if not matching_source:
                        continue
                matches += 1
                if return_format == self.RET_FORMAT_DATASETS:
                    yield ds
                elif return_format == self.RET_FORMAT_PRODUCT_GROUPED:
                    product_results.append(ds)
            if return_format == self.RET_FORMAT_PRODUCT_GROUPED and product_results:
                yield (product_results, product)

    def _search_flat(
            self,
            limit: int | None = None,
            source_filter: Mapping[str, QueryField] | None = None,
            archived: bool | None = False,
            **query: QueryField
    ) -> Iterable[Dataset]:
        return cast(Iterable[Dataset], self._search(
                    return_format=self.RET_FORMAT_DATASETS,
                    limit=limit,
                    source_filter=source_filter,
                    archived=archived,
                    **query)
        )

    def _search_grouped(
            self,
            limit: int | None = None,
            source_filter: Mapping[str, QueryField] | None = None,
            archived: bool | None = False,
            **query: QueryField
    ) -> Iterable[tuple[Iterable[Dataset], Product]]:
        return cast(Iterable[tuple[Iterable[Dataset], Product]], self._search(
                    return_format=self.RET_FORMAT_PRODUCT_GROUPED,
                    limit=limit,
                    source_filter=source_filter,
                    archived=archived,
                    **query)
        )

    def _get_prod_queries(self, **query: QueryField) -> Iterable[tuple[Mapping[str, QueryField], Product]]:
        return ((q, product) for product, q in self._index.products.search_robust(**query))

    @deprecat(
        deprecated_args={
            "source_filter": {
                "reason": "Filtering by source metadata is deprecated and will be removed in future.",
                "version": "1.9.0",
                "category": ODC2DeprecationWarning

            }
        }
    )
    def search(self,
               limit: int | None = None,
               source_filter: Mapping[str, QueryField] | None = None,
               archived: bool | None = False, order_by: Iterable[Any] | None = None,
               **query: QueryField) -> Iterable[Dataset]:
        if order_by:
            raise NotImplementedError("order_by argument is not currently supported by the memory index driver.")
        return cast(
            Iterable[Dataset],
            self._search_flat(limit=limit, source_filter=source_filter, archived=archived, **query)
        )

    def search_by_product(self,
                          archived: bool | None = False,
                          **query: QueryField) -> Iterable[tuple[Iterable[Dataset], Product]]:
        return self._search_grouped(archived=archived, **query)  # type: ignore[arg-type]

    def search_returning(self,
                         field_names: Iterable[str] | None = None,
                         custom_offsets: Mapping[str, Offset] | None = None,
                         limit: int | None = None,
                         archived: bool | None = False,
                         order_by: Iterable[Any] | None = None,
                         **query: QueryField) -> Iterable[tuple]:
        if "geopolygon" in query:
            raise NotImplementedError("Spatial search index API not supported by this index.")
        if order_by:
            raise NotImplementedError("order_by argument is not currently supported by the memory index driver.")
        # Note that this implementation relies on dictionaries being ordered by insertion - this has been the case
        # since Py3.6, and officially guaranteed behaviour since Py3.7.
        if field_names is None and custom_offsets is None:
            field_name_d = {f: None for f in self._index.products.get_field_names()}
        elif field_names:
            field_name_d = {f: None for f in field_names}
        else:
            field_name_d = {}

        if custom_offsets:
            custom_fields = build_custom_fields(custom_offsets)
            for name in custom_fields:
                field_name_d[name] = None
        else:
            custom_fields = {}

        #    Typing note: mypy can't handle dynamically created namedtuples
        result_type = namedtuple('search_result', field_name_d.keys())  # type: ignore[misc]
        for ds in self.search(limit=limit, archived=archived, **query):  # type: ignore[arg-type]
            ds_fields = get_dataset_fields(ds.metadata_type.definition)
            ds_fields.update(custom_fields)
            result_vals = {
                fn: ds_fields[fn].extract(ds.metadata_doc) if fn in ds_fields else None
                for fn in field_name_d.keys()
            }
            yield result_type(**result_vals)

    def count(self, archived: bool | None = False, **query: QueryField) -> int:
        return len(list(self.search(archived=archived, **query)))  # type: ignore[arg-type]

    def count_by_product(self, archived: bool | None = False, **query: QueryField) -> Iterable[tuple[Product, int]]:
        for datasets, prod in self.search_by_product(archived=archived, **query):
            yield (prod, len(list(datasets)))

    def count_by_product_through_time(self,
                                      period: str,
                                      archived: bool | None = False,
                                      **query: QueryField
    ) -> Iterable[
        tuple[
            Product,
            Iterable[
                tuple[Range, int]
            ]
        ]
    ]:
        return self._product_period_count(period, archived=archived, **query)  # type: ignore[arg-type]

    def _expand_period(
            self,
            period: str,
            begin: datetime.datetime,
            end: datetime.datetime
    ) -> Iterable[Range]:
        begin = tz_aware(begin)
        end = tz_aware(end)
        match = re.match(r'(?P<precision>[0-9]+) (?P<unit>day|month|week|year)', period)
        if not match:
            raise ValueError('Invalid period string. Must specify a number of days, weeks, months or years')
        precision = int(match.group("precision"))
        if precision <= 0:
            raise ValueError('Invalid period string. Must specify a natural number of days, weeks, months or years')
        unit = match.group("unit")

        def next_period(prev: datetime.datetime) -> datetime.datetime:
            if unit == 'day':
                return prev + datetime.timedelta(days=precision)
            elif unit == 'week':
                return prev + datetime.timedelta(days=precision * 7)
            elif unit == 'year':
                return datetime.datetime(
                    prev.year + precision,
                    prev.month,
                    prev.day,
                    prev.hour,
                    prev.minute,
                    prev.second,
                    tzinfo=prev.tzinfo
                )
            # unit == month
            year = prev.year
            month = prev.month
            month += precision
            while month > 12:
                month -= 12
                year += 1
            day = prev.day
            while True:
                try:
                    return datetime.datetime(
                        year,
                        month,
                        day,
                        prev.hour,
                        prev.minute,
                        prev.second,
                        tzinfo=prev.tzinfo
                    )
                except ValueError:
                    day -= 1
        period_start = begin
        while period_start < end:
            period_end = next_period(period_start)
            yield Range(begin=period_start, end=period_end)
            period_start = period_end

    def _product_period_count(
            self,
            period: str,
            single_product_only: bool = False,
            archived: bool | None = False,
            **query: QueryField
    ) -> Iterable[
        tuple[
            Product,
            Iterable[
                tuple[Range, int]
            ],
        ]
    ]:
        YieldType = tuple[Product, Iterable[tuple[Range, int]]]  # noqa: N806
        query = dict(query)
        try:
            start, end = cast(Range, query.pop('time'))
        except KeyError:
            raise ValueError('Must specify "time" range in period-counting query')
        periods = self._expand_period(period, start, end)
        last_product: YieldType | None = None
        for dss, product in self._search_grouped(archived=archived, **query):  # type: ignore[arg-type]
            if last_product and single_product_only:
                raise ValueError(f"Multiple products match single query search: {repr(query)}")
            if last_product:
                yield cast(YieldType, last_product)
            period_counts = []
            for p in periods:
                count = 0
                for ds in dss:
                    if ranges_overlap(cast(Range, ds.time), p):
                        count += 1
                period_counts.append((p, count))
            retval = (product, period_counts)
            if last_product is not None:
                yield retval
            last_product = retval

        if last_product is None:
            raise ValueError(f"No products match search terms: {repr(query)}")
        else:
            yield last_product

    def count_product_through_time(
            self,
            period: str,
            archived: bool | None = False,
            **query: QueryField
    ) -> Iterable[tuple[Range, int]]:
        return list(self._product_period_count(period, archived=archived, single_product_only=True, **query))[0][1]

    @deprecat(
        reason="This method is deprecated and will be removed in 2.0.  "
               "Consider migrating to search_returning()",
        version="1.9.0",
        category=ODC2DeprecationWarning
    )
    def search_summaries(self, **query: QueryField) -> Iterable[Mapping[str, Any]]:
        def make_summary(ds: Dataset) -> Mapping[str, Any]:
            fields = ds.metadata_type.dataset_fields
            return {
                field_name: field.extract(ds.metadata_doc)   # type: ignore[attr-defined]
                for field_name, field in fields.items()
            }
        for ds in self.search(**query):  # type: ignore[arg-type]
            yield make_summary(ds)

    def spatial_extent(self, ids, crs=None):
        return None

    def temporal_extent(self, ids: Iterable[DSID]) -> tuple[datetime.datetime, datetime.datetime]:
        min_time: datetime.datetime | None = None
        max_time: datetime.datetime | None = None
        if len(list(ids)) == 0:
            raise ValueError("no dataset ids provided")
        for dsid in ids:
            ds = self.get_unsafe(dsid)
            time_fld = ds.product.metadata_type.dataset_fields["time"]
            dsmin, dsmax = time_fld.extract(ds.metadata_doc)  # type: ignore[attr-defined]
            if dsmax is None and dsmin is None:
                continue
            elif dsmin is None:
                dsmin = dsmax
            elif dsmax is None:
                dsmax = dsmin
            if min_time is None or dsmin < min_time:
                min_time = dsmin
            if max_time is None or dsmax > max_time:
                max_time = dsmax
        return (cast(datetime.datetime, min_time), cast(datetime.datetime, max_time))

    # pylint: disable=redefined-outer-name
    def search_returning_datasets_light(
            self,
            field_names: tuple[str, ...],
            custom_offsets: Mapping[str, Offset] | None = None,
            limit: int | None = None,
            archived: bool | None = False,
            **query: QueryField
    ) -> Iterable[tuple]:
        if custom_offsets:
            custom_fields = build_custom_fields(custom_offsets)
        else:
            custom_fields = {}

        def make_ds_light(ds: Dataset) -> tuple:
            fields = {
                fname: ds.metadata_type.dataset_fields[fname]
                for fname in field_names
            }
            fields.update(custom_fields)
            #   Typing note: mypy cannot handle dynamically created namedtuples
            result_type = namedtuple('DatasetLight', list(fields.keys()))  # type: ignore[misc]
            if 'grid_spatial' in fields:
                class DatasetLight(result_type, DatasetSpatialMixin):
                    pass
            else:
                class DatasetLight(result_type):  # type: ignore[no-redef]
                    __slots__ = ()
            fld_vals = {
                fname: field.extract(ds.metadata_doc)  # type: ignore[attr-defined]
                for fname, field in fields.items()
            }
            return DatasetLight(**fld_vals)
        for ds in self.search(limit=limit, archived=archived, **query):  # type: ignore[arg-type]
            yield make_ds_light(ds)

    def clone(self, orig: Dataset, for_save=False, lookup_locations=True) -> Dataset:
        if for_save:
            uris = []
        elif lookup_locations:
            uris = self._locations[orig.id].copy()
        elif orig.uris:
            uris = orig.uris.copy()
        else:
            uris = []
        if len(uris) == 1:
            kwargs: dict[str, Any] = {"uri": uris[0]}
        elif len(uris) > 1:
            kwargs = {"uris": uris}
        else:
            kwargs = {}
        return Dataset(
            product=self._index.products.clone(orig.product),
            metadata_doc=jsonify_document(orig.metadata_doc_without_lineage()),
            indexed_by="user" if for_save and orig.indexed_by is None else orig.indexed_by,
            indexed_time=datetime.datetime.now() if for_save and orig.indexed_time is None else orig.indexed_time,
            archived_time=None if for_save else orig.archived_time,
            **kwargs
        )

    # Lineage methods need to be implemented on the dataset resource as that is where the relevant indexes
    # currently live.
    def _get_all_lineage(self) -> Iterable[LineageRelation]:
        for derived_id, sources in self._derived_from.items():
            for classifier, source_id in sources.items():
                yield LineageRelation(
                    derived_id=derived_id,
                    source_id=source_id,
                    classifier=classifier
                )

    def _add_lineage_batch(self, batch_rels: Iterable[LineageRelation]) -> BatchStatus:
        b_added = 0
        b_skipped = 0
        b_started = monotonic()
        for rel in batch_rels:
            if rel.derived_id in self._derived_from:
                if (rel.classifier in self._derived_from[rel.derived_id]
                        and self._derived_from[rel.derived_id][rel.classifier] != rel.source_id):
                    b_skipped += 1
                    continue
                else:
                    self._derived_from[rel.derived_id][rel.classifier] = rel.source_id
                    b_added += 1
            else:
                self._derived_from[rel.derived_id] = {rel.classifier: rel.source_id}
                b_added += 1

            if rel.source_id in self._derivations:
                self._derivations[rel.source_id][rel.classifier] = rel.derived_id
            else:
                self._derivations[rel.source_id] = {rel.classifier: rel.derived_id}

        return BatchStatus(b_added, b_skipped, monotonic()-b_started)


class LineageResource(NoLineageResource):
    """
    Minimal implementation as does not support external lineage.
    Lineage indexes live in the Dataset resource, so thin wrapper around that.
    """
    def get_all_lineage(self, batch_size: int = 1000) -> Iterable[LineageRelation]:
        return self._index.datasets._get_all_lineage()

    def _add_batch(self, batch_rels: Iterable[LineageRelation]) -> BatchStatus:
        return self._index.datasets._add_lineage_batch(batch_rels)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from unittest.mock import MagicMock

import xarray as xr
import numpy as np
import datetime
from uuid import UUID
from types import SimpleNamespace
import pytest

from datacube.api.query import GroupBy
from datacube.api.core import _calculate_chunk_sizes, output_geobox
from datacube import Datacube
from datacube.testutils.geom import AlbersGS
from datacube.testutils import mk_sample_dataset, suppress_deprecations


def test_grouping_datasets():
    def group_func(d):
        return d.time
    dimension = 'time'
    units = None
    datasets = [
        SimpleNamespace(time=datetime.datetime(2016, 1, 1), value='foo', id=UUID(int=10)),
        SimpleNamespace(time=datetime.datetime(2016, 2, 1), value='bar', id=UUID(int=1)),
        SimpleNamespace(time=datetime.datetime(2016, 1, 1), value='flim', id=UUID(int=9)),
    ]

    group_by = GroupBy(group_func, dimension, units, sort_key=group_func)
    grouped = Datacube.group_datasets(datasets, group_by)
    dss = grouped.isel(time=0).values[()]
    assert isinstance(dss, tuple)
    assert len(dss) == 2
    assert [ds.value for ds in dss] == ['flim', 'foo']

    dss = grouped.isel(time=1).values[()]
    assert isinstance(dss, tuple)
    assert len(dss) == 1
    assert [ds.value for ds in dss] == ['bar']

    assert str(grouped.time.dtype) == 'datetime64[ns]'
    assert grouped.loc['2016-01-01':'2016-01-15']


def test_group_datasets_by_time():
    bands = [dict(name='a')]
    # Same time instant but one explicitly marked as UTC
    ds1 = mk_sample_dataset(bands, timestamp="2019-01-01T23:24:00Z")
    ds2 = mk_sample_dataset(bands, timestamp="2019-01-01T23:24:00")
    # Same "time" but in a different timezone, and actually later
    ds3 = mk_sample_dataset(bands, timestamp="2019-01-01T23:24:00-1")
    assert ds1.center_time.tzinfo is not None
    assert ds2.center_time.tzinfo is None
    assert ds3.center_time.tzinfo is not None

    xx = Datacube.group_datasets([ds1, ds2, ds3], 'time')
    assert xx.time.shape == (2,)
    assert len(xx.data[0]) == 2
    assert len(xx.data[1]) == 1


def test_grouped_datasets_should_be_in_consistent_order():
    datasets = [
        {'time': datetime.datetime(2016, 1, 1, 0, 1), 'value': 'foo'},
        {'time': datetime.datetime(2016, 1, 1, 0, 2), 'value': 'flim'},
        {'time': datetime.datetime(2016, 2, 1, 0, 1), 'value': 'bar'}
    ]

    grouped = _group_datasets_by_date(datasets)

    # Swap the two elements which get grouped together
    datasets[0], datasets[1] = datasets[1], datasets[0]
    grouped_2 = _group_datasets_by_date(datasets)

    assert len(grouped) == len(grouped_2) == 2
    assert all(grouped.values == grouped_2.values)


def _group_datasets_by_date(datasets):
    def group_func(d):
        return d['time'].date()

    def sort_key(d):
        return d['time']
    dimension = 'time'
    units = None

    group_by = GroupBy(group_func, dimension, units, sort_key)
    return Datacube.group_datasets(datasets, group_by)


def test_dask_chunks():
    coords = {'time': np.arange(10)}

    sources = xr.DataArray(coords['time'],
                           coords=coords,
                           dims=list(coords))
    geobox = AlbersGS.tile_geobox((0, 0))[:6, :7]

    assert geobox.dimensions == ('y', 'x')
    assert sources.dims == ('time',)

    assert _calculate_chunk_sizes(sources, geobox, {}) == ((1,), (6, 7))
    assert _calculate_chunk_sizes(sources, geobox, {'time': -1}) == ((10,), (6, 7))
    assert _calculate_chunk_sizes(sources, geobox, {'time': 'auto', 'x': 'auto'}) == ((1,), (6, 7))
    assert _calculate_chunk_sizes(sources, geobox, {'y': -1, 'x': 3}) == ((1,), (6, 3))
    assert _calculate_chunk_sizes(sources, geobox, {'y': 2, 'x': 3}) == ((1,), (2, 3))

    with pytest.raises(ValueError):
        _calculate_chunk_sizes(sources, geobox, {'x': "aouto"})

    with pytest.raises(KeyError):
        _calculate_chunk_sizes(sources, geobox, {'zz': 1})


def test_index_validation():
    index = MagicMock()
    with pytest.raises(ValueError) as e:
        dc = Datacube(index=index, config=["/a/path", "/a/nother/path"], env="prod", app="this_is_me", raw_config="{}")
    estr = str(e.value)
    assert "config,raw_config,app,env" in estr


def test_output_geobox():
    from odc.geo.geobox import GeoBox as ODCGeoGeoBox, CRS as ODCGeoCRS
    with suppress_deprecations():
        from datacube.utils.geometry import GeoBox as LegacyGeoGeoBox, CRS as LegacyCRS
    from odc.geo.xr import xr_zeros

    odc_gbox = ODCGeoGeoBox.from_bbox(
        (-2_000_000, -5_000_000, 2_250_000, -1_000_000),
        "epsg:3577",
        resolution=1000
    )
    legacy_gbox = LegacyGeoGeoBox(width=odc_gbox.width, height=odc_gbox.height,
                                  affine=odc_gbox.affine,
                                  crs=LegacyCRS(str(odc_gbox.crs)))

    assert legacy_gbox != odc_gbox

    xra = xr_zeros(odc_gbox, chunks=(1000, 1000))

    assert xra.odc.geobox == odc_gbox

    gbox = output_geobox(like=xra)
    assert gbox == odc_gbox

    gbox = output_geobox(like=legacy_gbox)
    assert gbox == odc_gbox
    assert isinstance(gbox.crs, ODCGeoCRS)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import logging
import datetime

from datacube.index.abstract import AbstractProductResource
from datacube.model import Product

from typing import Iterable

_LOG = logging.getLogger(__name__)


class ProductResource(AbstractProductResource):
    def add(self, product, allow_table_lock=False):
        raise NotImplementedError()

    def can_update(self, product, allow_unsafe_updates=False, allow_table_lock=False):
        raise NotImplementedError()

    def update(self, product: Product, allow_unsafe_updates=False, allow_table_lock=False):
        raise NotImplementedError()

    def delete(self, products: Iterable[Product], allow_delete_active: bool = False):
        raise NotImplementedError()

    def get_unsafe(self, id_):
        raise KeyError(id_)

    def get_by_name_unsafe(self, name):
        raise KeyError(name)

    def search_robust(self, **query):
        return []

    def search_by_metadata(self, metadata):
        return []

    def get_all(self) -> Iterable[Product]:
        return []

    def temporal_extent(self, product: str | Product) -> tuple[datetime.datetime, datetime.datetime]:
        raise KeyError(str(product))

    def spatial_extent(self, product, crs=None):
        raise KeyError(str(product))

    def most_recent_change(self, product: str | Product) -> datetime.datetime | None:
        raise KeyError(str(product))



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""Initial revision

Revision ID: 48fb71ece268
Revises:
Create Date: 2023-11-21 16:28:45.209473

"""
from typing import Sequence, Union


# revision identifiers, used by Alembic.
revision: str = '48fb71ece268'
down_revision: Union[str, None] = None
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None

# It is helpful for schema management to have an empty initial revision


def upgrade() -> None:
    pass


def downgrade() -> None:
    pass



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Core SQL schema settings.
"""

import logging
import os

from sqlalchemy import MetaData, inspect, text
from sqlalchemy.engine import Engine
from sqlalchemy.schema import CreateSchema
from sqlalchemy.sql.ddl import DropSchema
from alembic import command, config
from alembic.migration import MigrationContext
from alembic.script import ScriptDirectory
from alembic.runtime.environment import EnvironmentContext

from datacube.drivers.postgis.sql import (INSTALL_TRIGGER_SQL_TEMPLATE,
                                          SCHEMA_NAME, TYPES_INIT_SQL,
                                          UPDATE_TIMESTAMP_SQL,
                                          escape_pg_identifier)

USER_ROLES = ('odc_user', 'odc_manage', 'odc_admin')

SQL_NAMING_CONVENTIONS = {
    "ix": 'ix_%(column_0_label)s',
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(constraint_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
    # Other prefixes handled outside of sqlalchemy:
    # dix: dynamic-index, those indexes created automatically based on search field configuration.
    # tix: test-index, created by hand for testing, particularly in dev.
}

POSTGIS_DRIVER_DIR = os.path.dirname(__file__)

ALEMBIC_INI_LOCATION = os.path.join(POSTGIS_DRIVER_DIR, "alembic.ini")

METADATA = MetaData(naming_convention=SQL_NAMING_CONVENTIONS, schema=SCHEMA_NAME)

_LOG = logging.getLogger(__name__)


def install_timestamp_trigger(connection):
    from . import _schema
    TABLE_NAMES = [  # noqa: N806
        _schema.MetadataType.__tablename__,
        _schema.Product.__tablename__,
        _schema.Dataset.__tablename__,
    ]
    # Create trigger capture function
    connection.execute(text(UPDATE_TIMESTAMP_SQL))

    for name in TABLE_NAMES:
        connection.execute(text(INSTALL_TRIGGER_SQL_TEMPLATE.format(schema=SCHEMA_NAME, table=name)))


def schema_qualified(name):
    """
    >>> schema_qualified('dataset')
    'odc.dataset'
    """
    return '{}.{}'.format(SCHEMA_NAME, name)


def _get_quoted_connection_info(connection):
    db, user = connection.execute(text("select quote_ident(current_database()), quote_ident(current_user)")).fetchone()
    return db, user


def ensure_db(engine, with_permissions=True):
    """
    Initialise the db if needed.

    Ensures standard users exist.

    Create the schema if it doesn't exist.
    """
    is_new = not has_schema(engine)
    with engine.connect() as c:
        #  NB. Using default SQLA2.0 auto-begin commit-as-you-go behaviour
        quoted_db_name, quoted_user = _get_quoted_connection_info(c)

        _ensure_extension(c, 'POSTGIS')
        c.commit()

        if with_permissions:
            _LOG.info('Ensuring user roles.')
            _ensure_role(c, 'odc_user')
            _ensure_role(c, 'odc_manage', inherits_from='odc_user')
            _ensure_role(c, 'odc_admin', inherits_from='odc_manage', add_user=True)

            c.execute(text(f"""
            grant all on database {quoted_db_name} to odc_admin;
            """))
            c.commit()

        if is_new:
            sqla_txn = c.begin()
            if with_permissions:
                # Switch to 'odc_admin', so that all items are owned by them.
                c.execute(text('set role odc_admin'))
            _LOG.info('Creating schema.')
            c.execute(CreateSchema(SCHEMA_NAME))
            _LOG.info('Creating types.')
            c.execute(text(TYPES_INIT_SQL))
            from ._schema import orm_registry, ALL_STATIC_TABLES
            _LOG.info('Creating tables.')
            _LOG.info("Dataset indexes: %s", repr(orm_registry.metadata.tables["odc.dataset"].indexes))
            orm_registry.metadata.create_all(c, tables=ALL_STATIC_TABLES)
            _LOG.info("Creating triggers.")
            install_timestamp_trigger(c)
            sqla_txn.commit()
            if with_permissions:
                c.execute(text(f'set role {quoted_user}'))
            c.commit()
            # Stamp with latest Alembic revision
            alembic_cfg = config.Config(ALEMBIC_INI_LOCATION)
            alembic_cfg.attributes["connection"] = c
            command.stamp(alembic_cfg, "head")

        if with_permissions:
            _LOG.info('Adding role grants.')
            c.execute(text(f"""
            grant usage on schema {SCHEMA_NAME} to odc_user;
            grant select on all tables in schema {SCHEMA_NAME} to odc_user;

            grant insert on {SCHEMA_NAME}.dataset,
                            {SCHEMA_NAME}.dataset_lineage to odc_manage;
            grant usage, select on all sequences in schema {SCHEMA_NAME} to odc_manage;

            -- Manage allows deletion of types that have nothing written yet (admin needed to delete the data itself)
            grant insert, delete on {SCHEMA_NAME}.product,
                                    {SCHEMA_NAME}.metadata_type to odc_manage;
            -- Allow creation of indexes, views
            grant create on schema {SCHEMA_NAME} to odc_manage;
            """))
            c.commit()

    return is_new


def database_exists(engine):
    """
    Have they init'd this database?
    """
    return has_schema(engine)


def schema_is_latest(engine: Engine) -> bool:
    """
    Is the current schema up-to-date?

    This is run when a new connection is established to see if it's compatible.

    It should be runnable by unprivileged users. If it returns false, their
    connection will be rejected and they will be told to get an administrator
    to apply updates.

    See the ``update_schema()`` function below for actually applying the updates.
    """

    # No schema changes recently. Everything is perfect.

    cfg = config.Config(ALEMBIC_INI_LOCATION)
    scriptdir = ScriptDirectory.from_config(cfg)
    # NB this assumes a single unbranched migration branch
    # Get Head revision from Alembic environment
    with EnvironmentContext(cfg, scriptdir) as env_ctx:
        latest_rev = env_ctx.get_head_revision()
        # Get current revision from database
        with engine.connect() as conn:
            context = MigrationContext.configure(
                connection=conn,
                environment_context=env_ctx,
                opts={"version_table_schema": "odc"}
            )
            current_rev = context.get_current_revision()

    # Do they match?
    if latest_rev == current_rev:
        return True
    import warnings
    warnings.warn(f"Current Alembic schema revision is {current_rev} expected {latest_rev}")
    return False


def update_schema(engine: Engine):
    """
    Check and apply any missing schema changes to the database.

    This is run by an administrator.

    See the `schema_is_latest()` function above: this should apply updates
    that it requires.
    """
    cfg = config.Config(ALEMBIC_INI_LOCATION)
    with engine.begin() as conn:
        cfg.attributes["connection"] = conn
        print("Running upgrade")
        command.upgrade(cfg, "head")


def _ensure_extension(conn, extension_name="POSTGIS"):
    sql = text(f'create extension if not exists {extension_name}')
    conn.execute(sql)


def _ensure_role(conn, name, inherits_from=None, add_user=False, create_db=False):
    if has_role(conn, name):
        _LOG.debug('Role exists: %s', name)
        return

    sql = [
        'create role %s nologin inherit' % name,
        'createrole' if add_user else 'nocreaterole',
        'createdb' if create_db else 'nocreatedb'
    ]
    if inherits_from:
        sql.append('in role ' + inherits_from)
    conn.execute(text(' '.join(sql)))


def grant_role(conn, role, users):
    if role not in USER_ROLES:
        raise ValueError('Unknown role %r. Expected one of %r' % (role, USER_ROLES))

    users = [escape_pg_identifier(conn, user) for user in users]
    conn.execute(text('revoke {roles} from {users}'.format(users=', '.join(users), roles=', '.join(USER_ROLES))))
    conn.execute(text('grant {role} to {users}'.format(users=', '.join(users), role=role)))


def has_role(conn, role_name):
    return bool(
        conn.execute(text(f"SELECT rolname FROM pg_roles WHERE rolname='{role_name}'")).fetchall()
    )


def has_schema(engine):
    inspector = inspect(engine)
    return SCHEMA_NAME in inspector.get_schema_names()


def drop_db(connection):
    # if_exists parameter seems to not be working in SQLA1.4?
    if has_schema(connection.engine):
        connection.execute(DropSchema(SCHEMA_NAME, cascade=True, if_exists=True))


def to_pg_role(role):
    """
    Convert a role name to a name for use in PostgreSQL

    There is a short list of valid ODC role names, and they are given
    a prefix inside of PostgreSQL.

    Why are we even doing this? Can't we use the same names internally and externally?

    >>> to_pg_role('user')
    'odc_user'
    >>> to_pg_role('fake')
    Traceback (most recent call last):
    ...
    ValueError: Unknown role 'fake'. Expected one of ...
    """
    pg_role = 'odc_' + role.lower()
    if pg_role not in USER_ROLES:
        raise ValueError(
            'Unknown role %r. Expected one of %r' %
            (role, [r.split('_')[1] for r in USER_ROLES])
        )
    return pg_role


def from_pg_role(pg_role):
    """
    Convert a PostgreSQL role name back to an ODC name.

    >>> from_pg_role('odc_admin')
    'admin'
    >>> from_pg_role('fake')
    Traceback (most recent call last):
    ...
    ValueError: Not a pg role: 'fake'. Expected one of ...
    """
    if pg_role not in USER_ROLES:
        raise ValueError('Not a pg role: %r. Expected one of %r' % (pg_role, USER_ROLES))

    return pg_role.split('_')[1]



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import logging

import click
from click import echo, style
from sqlalchemy.exc import OperationalError

import datacube
from datacube.cfg import ODCEnvironment, psql_url_from_config
from datacube.index import index_connect
from datacube.drivers.postgres._connections import IndexSetupError
from datacube.ui import click as ui
from datacube.ui.click import cli, handle_exception

_LOG = logging.getLogger('datacube-system')


@cli.group(name='system', help='System commands')
def system():
    pass


@system.command('init', help='Initialise the database')
@click.option(
    '--default-types/--no-default-types', is_flag=True, default=True,
    help="Add default types? (default: true)"
)
@click.option(
    '--init-users/--no-init-users', is_flag=True, default=True,
    help="Include user roles and grants. (default: true)"
)
@click.option(
    '--recreate-views/--no-recreate-views', is_flag=True, default=True,
    help="Recreate dynamic views"
)
@click.option(
    '--rebuild/--no-rebuild', is_flag=True, default=False,
    help="Rebuild all dynamic fields (caution: slow)"
)
@click.option(
    '--lock-table/--no-lock-table', is_flag=True, default=False,
    help="Allow table to be locked (eg. while creating missing indexes)"
)
@ui.pass_index(expect_initialised=False)
def database_init(index, default_types, init_users, recreate_views, rebuild, lock_table):
    echo('Initialising database...')

    was_created = index.init_db(with_default_types=default_types,
                                with_permissions=init_users)

    if was_created:
        echo(style('Created.', bold=True))
    else:
        echo(style('Updated.', bold=True))

    echo('Checking indexes/views.')
    index.metadata_types.check_field_indexes(
        allow_table_lock=lock_table,
        rebuild_indexes=rebuild,
        rebuild_views=recreate_views or rebuild,
    )
    echo('Done.')


@system.command('check', help='Check and display current configuration')
@ui.pass_config
def check(cfg_env: ODCEnvironment):
    """
    Verify & view current configuration
    """

    def echo_field(name, value):
        echo('{:<15}'.format(name + ':') + style(str(value), bold=True))

    echo_field('Version', datacube.__version__)
    echo_field('Index Driver', cfg_env.index_driver)
    db_url = psql_url_from_config(cfg_env)
    echo_field('Database URL:', db_url)
    echo_field('Skip broken datasets:', cfg_env.skip_broken_datasets)
    echo_field('Datacube load limit:', cfg_env.dc_load_limit)

    echo()
    echo('Valid connection:\t', nl=False)
    try:
        index = index_connect(config_env=cfg_env)
        echo(style('YES', bold=True))
        if index.url_parts.username:
            for role, user, description in index.users.list_users():
                if user == index.url_parts.username:
                    echo('You have %s privileges.' % style(role.upper(), bold=True))
    except OperationalError as e:
        handle_exception('Error Connecting to Database: %s', e)
    except IndexSetupError as e:
        handle_exception('Database not initialised: %s', e)


@system.command('clone', help='Clone an existing ODC index into a new, empty index')
@click.option('--batch-size',
              help='Size of batches for bulk-adding to the new index',
              type=int,
              default=1000)
@click.option(
    '--skip-lineage/--no-skip-lineage', is_flag=True, default=False,
    help="Do not load lineage data where possible. (default: false - i.e. do not skip lineage)"
)
@click.option(
    '--lineage-only/--no-lineage-only', is_flag=True, default=False,
    help="Clone lineage data only. (default: false)"
)
@click.argument('source-env', type=str, nargs=1)
@ui.pass_config
def clone(env: ODCEnvironment, batch_size: int, skip_lineage: bool, lineage_only: bool, source_env: str):
    if skip_lineage and lineage_only:
        echo("Cannot set both lineage-only and skip-lineage")
        exit(1)
    try:
        destination_index = index_connect(env, validate_connection=True)
    except OperationalError as e:
        handle_exception('Error Connecting to Destination Database: %s', e)
        exit(1)
    except IndexSetupError as e:
        handle_exception('Destination database not initialised: %s', e)
        exit(1)

    try:
        source_config = env._cfg[source_env]
    except KeyError:
        raise click.ClickException("No datacube config found for '{}'".format(source_env))
        exit(1)

    try:
        src_index = index_connect(source_config, validate_connection=True)
    except OperationalError as e:
        handle_exception('Error Connecting to Source Database: %s', e)
        exit(1)
    except IndexSetupError as e:
        handle_exception('Source database not initialised: %s', e)
        exit(1)
    # Any errors will have be logged.
    destination_index.clone(src_index, batch_size=batch_size, skip_lineage=skip_lineage, lineage_only=lineage_only)
    exit(0)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
odc.loader based load.

separate file to reduce formatting issues.

"""

from __future__ import annotations

from datetime import datetime
from types import SimpleNamespace
from typing import Sequence

import xarray as xr
from odc.geo.geobox import GeoBox, GeoboxTiles
from odc.loader import (
    FixedCoord,
    RasterBandMetadata,
    RasterGroupMetadata,
    RasterLoadParams,
    RasterSource,
    chunked_load,
    reader_driver,
    resolve_chunk_shape,
)
from odc.loader.types import ReaderDriverSpec

from ..model import Dataset, ExtraDimensions, Measurement
from . import BandInfo


def ds_geobox(ds: Dataset, **kw) -> GeoBox | None:
    from ..testutils.io import eo3_geobox

    try:
        return eo3_geobox(ds, **kw)
    except ValueError:
        return None


def _extract_coords(extra_dims: ExtraDimensions) -> list[FixedCoord]:
    coords = extra_dims.dims

    return [
        FixedCoord(
            name=k,
            values=d["values"],
            dtype=str(d.get("dtype", "float32")),
            dim=k,
            units=d.get("units", None),
        )
        for k, d in coords.items()
    ]


def driver_based_load(
    driver: ReaderDriverSpec,
    sources: xr.DataArray,
    geobox: GeoBox,
    measurements: Sequence[Measurement],
    dask_chunks=None,
    skip_broken_datasets=False,
    progress_cbk=None,
    extra_dims: ExtraDimensions | None = None,
    patch_url=None,
):
    fail_on_error = not skip_broken_datasets

    if extra_dims is None:
        extra_coords = []
    else:
        extra_coords = _extract_coords(extra_dims)

    tss = [
        datetime.fromtimestamp(float(ts) * 1e-9)
        for ts in sources.coords["time"].data.ravel()
    ]
    band_query: list[str] = [m.name for m in measurements]
    template = RasterGroupMetadata(
        bands={
            (m.name, 1): RasterBandMetadata(
                m.dtype, m.nodata, m.units, dims=tuple(m.get("dims", ()))
            )
            for m in measurements
        },
        aliases={name: [(name, 1)] for name in band_query},
        extra_dims={coord.dim: len(coord.values) for coord in extra_coords},
        extra_coords=extra_coords,
    )

    load_cfg = {
        m.name: RasterLoadParams(
            m.dtype,
            m.nodata,
            resampling=m.get("resampling", "nearest"),
            fail_on_error=fail_on_error,
            dims=tuple(m.get("dims", ())),
        )
        for m in measurements
    }

    chunks = dask_chunks

    if chunks is not None:
        chunk_shape = resolve_chunk_shape(
            len(tss), geobox, chunks, "float32", cfg=load_cfg
        )
    else:
        chunk_shape = (1, 2048, 2048)

    gbt = GeoboxTiles(geobox, (chunk_shape[1], chunk_shape[2]))

    tyx_bins: dict[tuple[int, int, int], list[int]] = {}
    srcs = []

    if patch_url is None:
        patch_url = lambda x: x  # noqa: E731

    def _dss():
        for tidx, dss in enumerate(sources.data):
            for ds in dss:
                yield tidx, ds

    def _ds_extract(ds: Dataset) -> dict[str, RasterSource]:
        out = {}
        for n in band_query:
            bi = BandInfo(ds, n)
            band_idx = bi.band if bi.band is not None else 1

            if bi.dims is not None and len(bi.dims) > 2:
                band_idx = 0  # 0 indicates extra dims

            out[n] = RasterSource(
                patch_url(bi.uri),
                band=band_idx,
                subdataset=bi.layer,
                geobox=ds_geobox(ds, band=n),
                meta=template.bands[(n, band_idx or 1)],
                driver_data=bi.driver_data,
            )

        return out

    for tidx, ds in _dss():
        srcs.append(_ds_extract(ds))
        for iy, ix in gbt.tiles(ds.extent):
            tyx_bins.setdefault((tidx, iy, ix), []).append(len(srcs) - 1)

    if driver == "kk-debug":
        return SimpleNamespace(
            load_cfg=load_cfg,
            template=template,
            srcs=srcs,
            tyx_bins=tyx_bins,
            gbt=gbt,
            tss=tss,
        )

    rdr = reader_driver(driver)

    return chunked_load(
        load_cfg,
        template,
        srcs,
        tyx_bins=tyx_bins,
        gbt=gbt,
        tss=tss,
        env=rdr.capture_env(),
        rdr=rdr,
        chunks=dask_chunks,
        progress=progress_cbk,
    )



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0


def test_index_clone(index_pair_populated_empty):
    pop_idx, empty_idx = index_pair_populated_empty
    assert list(empty_idx.products.get_all()) == []
    results = empty_idx.clone(pop_idx)
    assert "eo3" in results["metadata_types"].safe
    assert "ga_ls8c_ard_3" in results["products"].safe
    assert results["products"].skipped == 0
    assert results["datasets"].skipped == 0


def test_index_clone_small_batch(index_pair_populated_empty):
    pop_idx, empty_idx = index_pair_populated_empty
    assert list(empty_idx.products.get_all()) == []
    results = empty_idx.clone(pop_idx, batch_size=2)
    assert "eo3" in results["metadata_types"].safe
    assert "ga_ls8c_ard_3" in results["products"].safe
    assert results["products"].skipped == 0
    assert results["datasets"].skipped == 0


def test_index_clone_cli(cfg_env_pair, index_pair_populated_empty, clirunner):
    source_cfg, target_cfg = cfg_env_pair
    clirunner([
        '-E', target_cfg._name,
        'system', 'clone',
        '--lineage-only', '--skip-lineage',
        source_cfg._name
    ], skip_env=True, expect_success=False)
    clirunner([
        '-E', target_cfg._name,
        'system', 'clone',
        source_cfg._name
    ], skip_env=True, expect_success=True)


def test_index_clone_cli_small_batch(cfg_env_pair, index_pair_populated_empty, clirunner):
    source_cfg, target_cfg = cfg_env_pair
    clirunner([
        '-E', target_cfg._name,
        'system', 'clone',
        '--batch-size', '2',
        source_cfg._name
    ], skip_env=True, expect_success=True)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
User configuration.
"""

import warnings
warnings.warn("The old datacube.config  is no longer supported.  Please use the new datacube.cfg library")



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""Make 'added' and 'archived' columns indexed

Revision ID: 610f32dca3cb
Revises: 18afe14f2816
Create Date: 2024-08-08 04:46:55.465365

"""
from typing import Sequence, Union

from alembic import op


# revision identifiers, used by Alembic.
revision: str = '610f32dca3cb'
down_revision: Union[str, None] = '18afe14f2816'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.create_index("ix_dataset_added", "dataset", ["added"], schema="odc", if_not_exists=True)
    op.create_index("ix_dataset_archived", "dataset", ["archived"], schema="odc", if_not_exists=True)


def downgrade() -> None:
    op.drop_index("ix_dataset_added", schema="odc")
    op.drop_index("ix_dataset_archived", schema="odc")



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import logging
from abc import ABC, abstractmethod
from time import monotonic
from typing import Mapping, Iterable
from uuid import UUID

from datacube.model import LineageTree, LineageDirection, LineageRelation
from datacube.model.lineage import LineageRelations
from datacube.utils import report_to_user

from ._types import DSID, BatchStatus

_LOG = logging.getLogger(__name__)


class AbstractLineageResource(ABC):
    """
    Abstract base class for the Lineage portion of an index api.

    All LineageResource implementations should inherit from this base class.

    Note that this is a "new" resource only supported by new index drivers with `supports_external_lineage`
    set to True.  If a driver does NOT support external lineage, it can use extend the NoLineageResource class below,
    which is a minimal implementation of this resource that raises a NotImplementedError for all methods.

    However, any index driver that supports lineage must implement at least the get_all_lineage() and _add_batch()
    methods.
    """
    def __init__(self, index) -> None:
        self._index = index
        # THis is explicitly for indexes that do not support the External Lineage API.
        assert self._index.supports_external_lineage

    @abstractmethod
    def get_derived_tree(self, id_: DSID, max_depth: int = 0) -> LineageTree:
        """
        Extract a LineageTree from the index, with:
            - "id" at the root of the tree.
            - "derived" direction (i.e. datasets derived from id, datasets derived from
              datasets derived from id, etc.)
            - maximum depth as requested (default 0 = unlimited depth)

        Tree may be empty (i.e. just the root node) if no lineage for id is stored.

        :param id: the id of the dataset at the root of the returned tree
        :param max_depth: Maximum recursion depth.  Default/Zero = unlimited depth
        :return: A derived-direction Lineage tree with id at the root.
        """

    @abstractmethod
    def get_source_tree(self, id_: DSID, max_depth: int = 0) -> LineageTree:
        """
        Extract a LineageTree from the index, with:
            - "id" at the root of the tree.
            - "source" direction (i.e. datasets id was derived from, the dataset ids THEY were derived from, etc.)
            - maximum depth as requested (default 0 = unlimited depth)

        Tree may be empty (i.e. just the root node) if no lineage for id is stored.

        :param id: the id of the dataset at the root of the returned tree
        :param max_depth: Maximum recursion depth.  Default/Zero = unlimited depth
        :return: A source-direction Lineage tree with id at the root.
        """

    @abstractmethod
    def merge(self, rels: LineageRelations, allow_updates: bool = False, validate_only: bool = False) -> None:
        """
        Merge an entire LineageRelations collection into the database.

        :param rels: The LineageRelations collection to merge.
        :param allow_updates: If False and the merging rels would require index updates,
                              then raise an InconsistentLineageException.
        :param validate_only: If True, do not actually merge the LineageRelations, just check for inconsistency.
                              allow_updates and validate_only cannot both be True
        """

    @abstractmethod
    def add(self, tree: LineageTree, max_depth: int = 0, allow_updates: bool = False) -> None:
        """
        Add or update a LineageTree into the Index.

        If the provided tree is inconsistent with lineage data already
        recorded in the database, by default a ValueError is raised,
        If replace is True, the provided tree is treated as authoritative
        and the database is updated to match.

        :param tree: The LineageTree to add to the index
        :param max_depth: Maximum recursion depth. Default/Zero = unlimited depth
        :param allow_updates: If False and the tree would require index updates to fully
                              add, then raise an InconsistentLineageException.
        """

    @abstractmethod
    def remove(self, id_: DSID, direction: LineageDirection, max_depth: int = 0) -> None:
        """
        Remove lineage information from the Index.

        Removes lineage relation data only. Home values not affected.

        :param id_: The Dataset ID to start removing lineage from.
        :param direction: The direction in which to remove lineage (from id_)
        :param max_depth: The maximum depth to which to remove lineage (0/default = no limit)
        """

    @abstractmethod
    def set_home(self, home: str, *args: DSID, allow_updates: bool = False) -> int:
        """
        Set the home for one or more dataset ids.

        :param home: The home string
        :param args: One or more dataset ids
        :param allow_updates: Allow datasets with existing homes to be updated.
        :returns: The number of records affected.  Between zero and len(args).
        """

    @abstractmethod
    def clear_home(self, *args: DSID, home: str | None = None) -> int:
        """
        Clear the home for one or more dataset ids, or all dataset ids that currently have
        a particular home value.

        :param args: One or more dataset ids
        :param home: The home string.  Supply home or args - not both.
        :returns: The number of home records deleted. Usually len(args).
        """

    @abstractmethod
    def get_homes(self, *args: DSID) -> Mapping[UUID, str]:
        """
        Obtain a dictionary mapping UUIDs to home strings for the passed in DSIDs.

        If a passed in DSID does not have a home set in the database, it will not
        be included in the returned mapping.  i.e. a database index with no homes
        recorded will always return an empty mapping.

        :param args: One or more dataset ids
        :return: Mapping of dataset ids to home strings.
        """

    @abstractmethod
    def get_all_lineage(self, batch_size: int = 1000) -> Iterable[LineageRelation]:
        """
        Perform a batch-read of all lineage relations (as used by index clone operation)
        and return as an iterable stream of LineageRelation objects.

        API Note: This API method is not finalised and may be subject to change.

        :param batch_size: The number of records to read from the database at a time.
        :return: An iterable stream of LineageRelation objects.
        """

    @abstractmethod
    def _add_batch(self, batch_rels: Iterable[LineageRelation]) -> BatchStatus:
        """
        Add a single "batch" of LineageRelation objects.

        No default implementation is provided

        API Note: This API method is not finalised and may be subject to change.

        :param batch_rels: An iterable of one batch's worth of LineageRelation objects to add
        :return: BatchStatus named tuple, with `safe` set to None.
        """

    def bulk_add(self, relations: Iterable[LineageRelation], batch_size: int = 1000) -> BatchStatus:
        """
        Add a group of LineageRelation objects in bulk.

        API Note: This API method is not finalised and may be subject to change.

        :param relations: An Iterable of LineageRelation objects (i.e. as returned by get_all_lineage)
        :param batch_size: Number of lineage relations to add per batch (default 1000)
        :return: BatchStatus named tuple, with `safe` set to None.
        """

        def increment_progress():
            report_to_user(".", progress_indicator=True)

        n_batches = 0
        n_in_batch = 0
        added = 0
        skipped = 0
        batch = []
        job_started = monotonic()
        for rel in relations:
            batch.append(rel)
            n_in_batch += 1
            if n_in_batch >= batch_size:
                batch_result = self._add_batch(batch)
                _LOG.info("Batch %d/%d datasets added in %.2fs: (%.2fdatasets/min)",
                          batch_result.completed,
                          n_in_batch,
                          batch_result.seconds_elapsed,
                          batch_result.completed * 60 / batch_result.seconds_elapsed)
                added += batch_result.completed
                skipped += batch_result.skipped
                batch = []
                n_in_batch = 0
                n_batches += 1
                increment_progress()
        if n_in_batch > 0:
            batch_result = self._add_batch(batch)
            added += batch_result.completed
            skipped += batch_result.skipped
            increment_progress()

        return BatchStatus(added, skipped, monotonic() - job_started)


class NoLineageResource(AbstractLineageResource):
    """
    Minimal implementation of AbstractLineageResource that raises "not implemented"
       for all methods.

    Index drivers that do not support lineage at all may use this implementation as is.

    Index drivers that support legacy lineage should extend this implementation and provide
    implementations of the get_all_lineage() and _add_batch() methods.
    """
    def __init__(self, index) -> None:
        self._index = index
        assert not self._index.supports_external_lineage

    def get_derived_tree(self, id: DSID, max_depth: int = 0) -> LineageTree:
        raise NotImplementedError()

    def get_source_tree(self, id: DSID, max_depth: int = 0) -> LineageTree:
        raise NotImplementedError()

    def add(self, tree: LineageTree, max_depth: int = 0, allow_updates: bool = False) -> None:
        raise NotImplementedError()

    def merge(self, rels: LineageRelations, allow_updates: bool = False, validate_only: bool = False) -> None:
        raise NotImplementedError()

    def remove(self, id_: DSID, direction: LineageDirection, max_depth: int = 0) -> None:
        raise NotImplementedError()

    def set_home(self, home: str, *args: DSID, allow_updates: bool = False) -> int:
        raise NotImplementedError()

    def clear_home(self, *args: DSID, home: str | None = None) -> int:
        raise NotImplementedError()

    def get_homes(self, *args: DSID) -> Mapping[UUID, str]:
        return {}

    def get_all_lineage(self, batch_size: int = 1000) -> Iterable[LineageRelation]:
        raise NotImplementedError()

    def _add_batch(self, batch_rels: Iterable[LineageRelation]) -> BatchStatus:
        raise NotImplementedError()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

import re
from typing import Any, TypeAlias, Callable

from .exceptions import ConfigException


# A raw configuration dictionary. A dictionary of dictionaries
ConfigDict: TypeAlias = dict[str, dict[str, Any]]
SemaphoreCallback = Callable[[], None]


def check_valid_env_name(name: str) -> None:
    """
    Enforce a valid ODC environment name.

    :param name: Candidate name
    :return:  None (raises ConfigException if invalid)
    """
    if not re.fullmatch(r"^[a-z][a-z0-9]*$", name):
        raise ConfigException(f'Environment names must consist of only lower case letters and numbers: {name}')
    if name.lower() == "all":
        raise ConfigException('Environments cannot be named "ALL"')


def check_valid_option(name: str) -> None:
    """
    Enforce a valid ODC config option name.

    :param name: Candidate name
    :return:  None (raises ConfigException if invalid)
    """
    if not re.fullmatch(r"^[a-z][a-z_]*$", name):
        raise ConfigException(
            f'Config option names must consist of only lower case letters, numbers and underscores: {name}'
        )


def smells_like_ini(cfg_text: str):
    """
    Does this file smell like an INI file?

    (Is the first non-whitespace character a '[' or ';'?)

    :param cfg_text: The contents of the file to be smelled
    :return: True if it smells like an ini file.
    """
    for line in cfg_text.split('\n'):
        line = line.strip()
        if not line:
            continue
        if line[0] in [";", "["]:
            return True
        else:
            return False
    # Empty file - parse as ini so we end up with the right high level structure at least
    return True



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

import datetime
from deprecat import deprecat

from datacube.migration import ODC2DeprecationWarning
from datacube.index.abstract import AbstractDatasetResource, DSID
from datacube.model import Dataset, Product
from typing import Iterable, Optional


class DatasetResource(AbstractDatasetResource):
    def __init__(self, index):
        super().__init__(index)

    def get_unsafe(self, id_: DSID, include_sources: bool = False, include_deriveds: bool = False, max_depth: int = 0):
        raise KeyError(id_)

    def bulk_get(self, ids):
        return []

    def get_derived(self, id_):
        return []

    def has(self, id_):
        return False

    def bulk_has(self, ids_):
        return [False for id_ in ids_]

    def add(self, dataset: Dataset,
            with_lineage: bool = True,
            archive_less_mature: Optional[int] = None) -> Dataset:
        raise NotImplementedError()

    def search_product_duplicates(self, product: Product, *args):
        return []

    def can_update(self, dataset, updates_allowed=None):
        raise NotImplementedError()

    def update(self, dataset: Dataset, updates_allowed=None, archive_less_mature=None):
        raise NotImplementedError()

    def archive(self, ids):
        raise NotImplementedError()

    def restore(self, ids):
        raise NotImplementedError()

    def purge(self, ids: Iterable[DSID], allow_delete_active: bool = False):
        raise NotImplementedError()

    def get_all_dataset_ids(self, archived: bool):
        return []

    @deprecat(
        reason="Multiple locations per dataset are now deprecated.  Please use the 'get_location' method.",
        version="1.9.0",
        category=ODC2DeprecationWarning
    )
    def get_locations(self, id_):
        return []

    def get_location(self, id_):
        return None

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
               "Archived locations may not be accessible in future releases.",
        version="1.9.0",
        category=ODC2DeprecationWarning
    )
    def get_archived_locations(self, id_):
        return []

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
               "Archived locations may not be accessible in future releases.",
        version="1.9.0",
        category=ODC2DeprecationWarning
    )
    def get_archived_location_times(self, id_):
        return []

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
               "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning
    )
    def add_location(self, id_, uri):
        raise NotImplementedError()

    def get_datasets_for_location(self, uri, mode=None):
        return []

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
               "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning
    )
    def remove_location(self, id_, uri):
        raise NotImplementedError()

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
               "Archived locations may not be accessible in future releases. "
               "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning
    )
    def archive_location(self, id_, uri):
        raise NotImplementedError()

    @deprecat(
        reason="Multiple locations per dataset are now deprecated. "
               "Archived locations may not be restorable in future releases. "
               "Dataset location can be set or updated with the update() method.",
        version="1.9.0",
        category=ODC2DeprecationWarning
    )
    def restore_location(self, id_, uri):
        raise NotImplementedError()

    def search_by_metadata(self, metadata, archived=False):
        return []

    @deprecat(
        deprecated_args={
            "source_filter": {
                "reason": "Filtering by source metadata is deprecated and will be removed in future.",
                "version": "1.9.0",
                "category": ODC2DeprecationWarning

            }
        }
    )
    def search(self, limit=None, archived=False, order_by=None, **query):
        return []

    def search_by_product(self, archived=False, **query):
        return []

    def search_returning(self,
                         field_names=None, custom_offsets=None,
                         limit=None, archived=False, order_by=None,
                         **query):
        return []

    def count(self, archived=False, **query):
        return 0

    def count_by_product(self, archived=False, **query):
        return []

    def count_by_product_through_time(self, period, archived=False, **query):
        return []

    def count_product_through_time(self, period, archived=False, **query):
        return []

    @deprecat(
        reason="This method is deprecated and will be removed in 2.0.  "
               "Consider migrating to search_returning()",
        version="1.9.0",
        category=ODC2DeprecationWarning
    )
    def search_summaries(self, **query):
        return []

    def temporal_extent(self, ids: Iterable[DSID]) -> tuple[datetime.datetime, datetime.datetime]:
        raise KeyError(str(ids))

    # pylint: disable=redefined-outer-name
    def search_returning_datasets_light(self,
                                        field_names: tuple,
                                        custom_offsets=None, limit=None, archived=False, **query):
        return []

    def spatial_extent(self, ids=None, product=None, crs=None):
        return None



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Modules for the Storage and Access Query API
"""

from .core import Datacube, TerminateCurrentLoad

__all__ = (
    'Datacube',
    'TerminateCurrentLoad',
)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

from datacube import Datacube
from datacube.cfg import ODCConfig


def test_multiple_environment_config(tmpdir):
    raw_config = """
[DEFAULT]
db_username: test_user
index_driver: default

[default]
db_hostname: db.opendatacube.test

[testalt]
db_hostname: alt-db.opendatacube.test
    """

    cfg = ODCConfig(text=raw_config)
    cfg_env = cfg[None]
    assert cfg_env.db_hostname == 'db.opendatacube.test'
    alt_env = cfg['testalt']
    assert alt_env.db_hostname == 'alt-db.opendatacube.test'
    assert cfg_env.index_driver == 'default'

    # Make sure the correct config is passed through the API
    # Parsed config:
    db_url = f'postgresql://{cfg_env.db_username}@db.opendatacube.test:5432/datacube'
    alt_db_url = f'postgresql://{alt_env.db_username}@alt-db.opendatacube.test:5432/datacube'

    with Datacube(env=cfg_env, validate_connection=False) as dc:
        assert str(dc.index.url) == db_url

    # When none specified, default environment is loaded
    with Datacube(config=cfg, validate_connection=False) as dc:
        assert str(dc.index.url) == db_url
    # When specific environment is loaded
    with Datacube(config=cfg, env='testalt', validate_connection=False) as dc:
        assert str(dc.index.url) == alt_db_url



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Add geometric extensions to :class:`xarray.Dataset` and :class:`xarray.DataArray` for use
with Data Cube by Monkey Patching those classes.

This extension is reliant on an `xarray` object having a `.crs` property of type
:class:`odc.geo.crs.CRS`. This is used to inspect the spatial dimensions of the
:class:`Dataset` or :class:`DataArray`, and provide new attributes for accessing a
:class:`odc.geo.geobox.GeoBox`, affine transform and extent for the dataset as
`.geobox`, `.affine` and `.extent` respectively.

"""
import warnings
import xarray
from odc.geo import resxy_
from odc.geo.math import affine_from_axis
from odc.geo.xr import spatial_dims
from odc.geo._xr_interop import _xarray_geobox as _xr_geobox


def _xarray_affine_impl(obj):
    sdims = spatial_dims(obj, relaxed=True)
    if sdims is None:
        return None, None

    yy, xx = (obj[dim] for dim in sdims)
    fallback_res = (coord.attrs.get('resolution', None) for coord in (xx, yy))
    res_x, res_y = next(fallback_res), next(fallback_res)
    fallback_res = None if (res_x, res_y) == (None, None) else resxy_(res_x, res_y)
    return affine_from_axis(xx.values, yy.values, fallback_res), sdims


def _xarray_affine(obj):
    transform, _ = _xarray_affine_impl(obj)
    return transform


def _xarray_extent(obj):
    geobox = obj.odc.geobox
    return None if geobox is None else geobox.extent


def _xarray_geobox(obj):
    warnings.warn(
        'Geobox extraction logic has moved to odc-geo and the .geobox property is now deprecated.'
        'Please access via .odc.geobox instead.',
        DeprecationWarning,
        stacklevel=2)
    return _xr_geobox(obj)


xarray.Dataset.geobox = property(_xarray_geobox)    # type: ignore
xarray.Dataset.affine = property(_xarray_affine)    # type: ignore
xarray.Dataset.extent = property(_xarray_extent)    # type: ignore
xarray.DataArray.geobox = property(_xarray_geobox)  # type: ignore
xarray.DataArray.affine = property(_xarray_affine)  # type: ignore
xarray.DataArray.extent = property(_xarray_extent)  # type: ignore



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from abc import ABC, abstractmethod
from typing import Iterable


class AbstractUserResource(ABC):
    """
    Abstract base class for the User portion of an index api.

    All UserResource implementations should inherit from this base
    class and implement all abstract methods.

    (If a particular abstract method is not applicable for a particular implementation
    raise a NotImplementedError)
    """

    @abstractmethod
    def grant_role(self, role: str, *usernames: str) -> None:
        """
        Grant a role to users
        :param role: name of the database role
        :param usernames: usernames to grant the role to.
        """

    @abstractmethod
    def create_user(self,
                    username: str,
                    password: str,
                    role: str,
                    description: str | None = None) -> None:
        """
        Create a new user
        :param username: username of the new user
        :param password: password of the new user
        :param role: default role of the the new user
        :param description: optional description for the new user
        """

    @abstractmethod
    def delete_user(self,
                    *usernames: str
                   ) -> None:
        """
        Delete database users
        :param usernames: usernames of users to be deleted
        """

    @abstractmethod
    def list_users(self) -> Iterable[tuple[str, str, str | None]]:
        """
        List all database users
        :return: Iterable of (role, username, description) tuples
        """



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from unittest.mock import MagicMock

import pytest


def test_smells_like_ini():
    from datacube.cfg.utils import smells_like_ini
    assert smells_like_ini("""[an_ini_file]
key: value
other-key: 12

""")
    assert smells_like_ini("""
; This is an ini comment
[an_ini_file]
key: value
other-key: 12

""")
    assert not smells_like_ini("""# This is a YAML comment

spagoots:
  four: score
  and: 7
  years: ago
""")
    assert not smells_like_ini("""{
"spagoots":{
  "four": "score",
  "and": 7,
  "years": "ago",
  # no comments in json!
  }
}
""")
    assert smells_like_ini("""[
    Looks ini-ish but
    :::is:::
    GIBBER-ish ..;';'@#$%^@#$%^#$^&$
\n  \n    \n \n  \t  \t  \n   \n
aasdfer\\faw043[]]][""")

    # Pure white space
    assert smells_like_ini("   \n  \n    \n \n  \t  \t  \n   \n")


@pytest.fixture
def simple_valid_ini():
    return """[foo]
bar: bell
bat: 7
baz: luhrmann
[goo]
bar: beque
bat: 2
baz: mckenzie
"""


@pytest.fixture
def simple_valid_yaml():
    return """foo:
    bar: bell
    bat: 7
    baz: luhrmann
goo:
    bar: beque
    bat: 2
    baz: mckenzie
"""


def test_parse_text(simple_valid_ini, simple_valid_yaml):
    from datacube.cfg import ConfigException, parse_text, CfgFormat
    ini = parse_text(simple_valid_ini)
    yaml = parse_text(simple_valid_yaml)
    assert ini["foo"]["bar"] == yaml["foo"]["bar"]
    assert ini["goo"]["baz"] == yaml["goo"]["baz"]
    assert int(ini["foo"]["bat"]) == int(yaml["foo"]["bat"])

    with pytest.raises(ConfigException) as e:
        ini_as_yaml = parse_text(simple_valid_ini, fmt=CfgFormat.YAML)

    with pytest.raises(ConfigException) as e:
        yaml_as_ini = parse_text(simple_valid_yaml, fmt=CfgFormat.INI)


@pytest.fixture
def single_env_config():
    return """# Simple single environment config
experimental:
   index_driver: postgis
   db_url: postgresql://foo:bar@server.subdomain.domain/mytestdb
   db_iam_authentication: yes
"""


@pytest.fixture
def simple_config():
    return """# Simple but thorough test config
default:
   alias: legacy
postgres:
   alias: legacy
legacy:
   index_driver: default
   db_username: foo
   db_password: bar
   db_hostname: server.subdomain.domain
   db_port: 5433
   db_database: mytestdb
   db_connection_timeout: 20
experimental:
   index_driver: postgis
   db_url: postgresql://foo:bar@server.subdomain.domain/mytestdb
   db_iam_authentication: yes
postgis:
   alias: experimental
memory:
   index_driver: memory
   db_url: '@nota?valid:url//foo&bar%%%'
exp2:
   index_driver: postgis
   db_url: postgresql://foo:bar@server.subdomain.domain/mytestdb
   db_database: not_read
   db_port: ignored
   db_iam_authentication: yes
   db_iam_timeout: 300
"""


@pytest.fixture
def simple_dict():
    return {
        "default": {"alias": "legacy"},
        "postgres": {"alias": "legacy"},
        "legacy": {
            "index_driver": "default",
            "db_username": "foo",
            "db_password": "bar",
            "db_hostname": "server.subdomain.domain",
            "db_port": 5433,
            "db_database": "mytestdb",
            "db_connection_timeout": 20
        },
        "experimental": {
            "index_driver": "postgis",
            "db_url": "postgresql://foo:bar@server.subdomain.domain/mytestdb",
            "db_iam_authentication": "yes"
        },
        "postgis": {"alias": "experimental"},
        "memory": {
            "index_driver": "memory",
            "db_url": "@nota?valid:url//foo&bar%%%"
        },
        "exp2": {
            "index_driver": "postgis",
            "db_url": "postgresql://foo:bar@server.subdomain.domain/mytestdb",
            "db_database": "not_read",
            "db_port": "ignored",
            "db_iam_authentication": "yes",
            "db_iam_timeout": 300
        }
    }


def test_invalid_env():
    from datacube.cfg import ODCConfig, ConfigException
    with pytest.raises(ConfigException):
        cfg = ODCConfig(text="""
default:
    alias: royale_avec_fromage
            """)
    with pytest.raises(ConfigException):
        cfg = ODCConfig(text="""
default:
    alias: legit00
non_legit:
    index_driver: null
        """)


def test_oldstyle_cfg():
    from datacube.cfg import ODCConfig
    with pytest.warns(UserWarning, match=r'default_environment.*no longer supported'):
        cfg = ODCConfig(text="""
        default:
            index_driver: memory
        env2:
            index_driver: memory
        env3:
            index_driver: memory
        user:
            default_environment: env3
        """)
        assert cfg[None]._name == 'default'


def test_invalid_option():
    from datacube.cfg import ODCOptionHandler, ConfigException
    mockenv = MagicMock()
    with pytest.raises(ConfigException):
        handler = ODCOptionHandler("NO_CAPS", mockenv)


def test_single_env(single_env_config):
    from datacube.cfg import ODCConfig
    cfg = ODCConfig(text=single_env_config)

    assert cfg['experimental'].index_driver == "postgis"
    assert cfg['experimental'].db_url == "postgresql://foo:bar@server.subdomain.domain/mytestdb"
    assert cfg['experimental'].db_username == "foo"
    with pytest.raises(AttributeError):
        assert cfg['experimental'].not_an_option
    assert cfg['experimental']['db_iam_authentication']
    assert cfg['experimental'].db_iam_timeout == 600
    assert cfg['experimental']['db_connection_timeout'] == 60


def assert_simple_aliases(cfg):
    assert cfg['default']._name == 'legacy'
    assert cfg['postgres']._name == 'legacy'
    assert cfg['legacy']._name == 'legacy'
    assert cfg['postgis']._name == 'experimental'
    assert cfg['experimental']._name == 'experimental'
    assert cfg['memory']._name == 'memory'
    assert cfg['exp2']._name == 'exp2'
    assert cfg['dynamic']._name == 'dynamic'
    assert cfg[None]._name == 'legacy'


def test_aliases(simple_config):
    from datacube.cfg import ODCConfig, ConfigException
    cfg = ODCConfig(text=simple_config)
    assert_simple_aliases(cfg)
    with pytest.raises(ConfigException) as e:
        cfg = ODCConfig(raw_dict={
            "default": {
                "alias": "main",
                "invalid_option": True
            },
            "main": {
                "index_driver": "memory"
            }
        })
    assert "invalid_option" in str(e.value)


def assert_simple_options(cfg):
    assert cfg['default']['index_driver'] == 'default'
    assert cfg['default'].db_username == 'foo'
    assert not cfg['default']['db_iam_authentication']
    with pytest.raises(KeyError):
        assert cfg['default']["db_iam_timeout"]

    assert cfg['exp2'].db_url == "postgresql://foo:bar@server.subdomain.domain/mytestdb"
    assert cfg['exp2'].db_username == "foo"
    with pytest.raises(AttributeError):
        assert cfg['exp2'].not_an_option
    assert cfg['exp2']['db_iam_authentication']
    assert cfg['exp2'].db_iam_timeout == 300
    assert cfg['exp2']['db_connection_timeout'] == 60


def test_options(simple_config):
    from datacube.cfg import ODCConfig
    cfg = ODCConfig(text=simple_config)
    assert_simple_options(cfg)


def test_rawdict(simple_dict):
    from datacube.cfg import ODCConfig
    cfg = ODCConfig(raw_dict=simple_dict)
    assert_simple_aliases(cfg)
    assert_simple_options(cfg)


def test_noenv_overrides_in_text(simple_config, monkeypatch):
    monkeypatch.setenv("ODC_LEGACY_DB_USERNAME", "bar")
    monkeypatch.setenv("ODC_EXPERIMENTAL_DB_USERNAME", "bar")
    from datacube.cfg import ODCConfig
    cfg = ODCConfig(text=simple_config)

    assert cfg["legacy"].db_username != 'bar'
    assert cfg["experimental"].db_username != "bar"


@pytest.fixture
def path_to_yaml_config():
    import os.path
    return os.path.join(os.path.dirname(__file__), "cfg", "simple_cfg.yaml")


@pytest.fixture
def path_to_ini_config():
    import os.path
    return os.path.join(os.path.dirname(__file__), "cfg", "simple_cfg.ini")


@pytest.fixture
def path_to_different_config():
    import os.path
    return os.path.join(os.path.dirname(__file__), "cfg", "different_cfg.yaml")


def test_yaml_from_path(path_to_yaml_config):
    from datacube.cfg import ODCConfig
    cfg = ODCConfig(paths=path_to_yaml_config)
    assert_simple_aliases(cfg)
    assert_simple_options(cfg)


def test_ini_from_path(path_to_ini_config):
    from datacube.cfg import ODCConfig
    cfg = ODCConfig(paths=path_to_ini_config)
    assert_simple_aliases(cfg)
    assert_simple_options(cfg)


def test_ini_from_paths(path_to_ini_config, path_to_yaml_config, path_to_different_config, monkeypatch):
    from datacube.cfg import ODCConfig, ConfigException
    cfg = ODCConfig(paths=[
        "/non/existent/path.yml",
        path_to_yaml_config,
        path_to_different_config,
        path_to_ini_config
    ])
    assert cfg[None]._name == 'legacy'

    cfg = ODCConfig(paths=[
        path_to_different_config,
        "/non/existent/path.yml",
        path_to_yaml_config,
    ])
    assert cfg[None]._name == 'experimental'

    with pytest.raises(ConfigException):
        cfg = ODCConfig(paths=[
            "/non/existent/path.yml",
            "/etc/shadow",  # A path that exists that we can't read
            "/another/nonexistent/path.yml"
        ])

    with monkeypatch.context() as mp:
        mp.setattr("datacube.cfg.cfg._DEFAULT_CONFIG_SEARCH_PATH", [
            "/non/existent/path.yml",
            "/etc/shadow",  # A path that exists that we can't read
            "/another/nonexistent/path.yml",
            path_to_yaml_config,
        ])
        cfg = ODCConfig()
        assert cfg[None]._name == 'legacy'

    with monkeypatch.context() as mp:
        mp.setattr("datacube.cfg.cfg._DEFAULT_CONFIG_SEARCH_PATH", [
            "/non/existent/path.yml",
            "/etc/shadow",  # A path that exists that we can't read
            "/another/nonexistent/path.yml",
        ])
        cfg = ODCConfig()
        assert cfg[None]._name == 'default'

    with monkeypatch.context() as mp:
        mp.setenv("ODC_CONFIG_PATH",
                  f"/non/existent/path.yml:{path_to_yaml_config}:{path_to_different_config}")
        cfg = ODCConfig()
        assert cfg[None]._name == 'legacy'

    with monkeypatch.context() as mp:
        mp.setenv("DATACUBE_CONFIG_PATH",
                  f"/non/existent/path.yml:{path_to_yaml_config}:{path_to_different_config}")
        cfg = ODCConfig()
        assert cfg[None]._name == 'legacy'


def test_envvar_overrides(path_to_yaml_config, monkeypatch):
    monkeypatch.setenv("ODC_LEGACY_DB_USERNAME", "bar")
    monkeypatch.setenv("ODC_EXPERIMENTAL_DB_USERNAME", "bar")
    monkeypatch.setenv("ODC_EXP2_DB_CONNECTION_TIMEOUT", "20")
    monkeypatch.setenv("DATACUBE_IAM_AUTHENTICATION", "yes")

    from datacube.cfg import ODCConfig
    cfg = ODCConfig(paths=path_to_yaml_config)
    assert cfg["legacy"].db_username == 'bar'
    assert cfg["legacy"].db_iam_authentication
    assert cfg["experimental"].db_iam_authentication
    assert cfg["exp2"].db_iam_authentication
    assert cfg["exp2"].db_connection_timeout == 20
    assert cfg["experimental"].db_username != 'bar'


def test_intopt_validation():
    from datacube.cfg import ODCConfig, ConfigException
    cfg = ODCConfig(text="""
env1:
    db_hostname: localhost
    db_port: seven
""")
    with pytest.raises(ConfigException):
        assert cfg["env1"].db_hostname == 'localhost'
    cfg = ODCConfig(text="""
env1:
    db_hostname: localhost
    db_port: -7
""")
    with pytest.raises(ConfigException):
        assert cfg["env1"].db_hostname == 'localhost'
    cfg = ODCConfig(text="""
env1:
    db_hostname: localhost
    db_port: 4425542239934
""")
    with pytest.raises(ConfigException):
        assert cfg["env1"].db_hostname == 'localhost'
    cfg = ODCConfig(text="""
env1:
    db_hostname: localhost
    db_port: 0
""")
    with pytest.raises(ConfigException):
        assert cfg["env1"].db_hostname == 'localhost'


def test_invalid_idx_driver():
    from datacube.cfg import ODCConfig, ConfigException
    cfg = ODCConfig(raw_dict={
        "default": {"alias": "foo"},
        "foo": {
            "index_driver": "phillips_head",
        }
    })
    with pytest.raises(ConfigException) as e:
        cfg["default"].index_driver
    estr = str(e.value)
    assert "Unknown index driver" in estr
    assert "phillips_head" in estr
    assert "legacy" in estr


def test_invalid_pg_url():
    from datacube.cfg import ODCConfig, ConfigException
    cfg = ODCConfig(raw_dict={
        "default": {"alias": "foo"},
        "foo": {
            "index_driver": "postgres",
            "db_url": "https://www.google.com"
        }
    })
    with pytest.raises(ConfigException):
        assert cfg["default"].index_driver == "postgres"


def test_pgurl_from_config(simple_dict):
    from datacube.cfg import ODCConfig, psql_url_from_config
    cfg = ODCConfig(raw_dict=simple_dict)
    assert psql_url_from_config(
        cfg["legacy"]
    ) == "postgresql://foo:bar@server.subdomain.domain:5433/mytestdb"
    assert psql_url_from_config(
        cfg["experimental"]
    ) == "postgresql://foo:bar@server.subdomain.domain/mytestdb"
    with pytest.raises(AttributeError):
        psql_url_from_config(
            cfg["memory"]
        )
    assert cfg["exp2"].db_url == "postgresql://foo:bar@server.subdomain.domain/mytestdb"
    assert cfg["exp2"].db_username == "foo"
    assert cfg["exp2"].db_password == "bar"
    assert cfg["exp2"].db_hostname == "server.subdomain.domain"
    assert not cfg["exp2"].db_port
    assert cfg["exp2"].db_database == "mytestdb"

    cfg = ODCConfig(raw_dict={
        "foo": {
            "db_hostname": "remotehost.local",
            "db_username": "penelope",
            "db_database": "mydb",
            "db_port": 5544,
        }
    })
    assert psql_url_from_config(
        cfg["foo"]
    ) == "postgresql://penelope@remotehost.local:5544/mydb"
    cfg = ODCConfig(raw_dict={
        "foo": {
            "db_hostname": "remotehost.local",
            "db_username": "penelope",
            "db_port": 5544,
        }
    })


def test_multiple_sourcetypes(simple_config, path_to_ini_config, simple_dict):
    from datacube.cfg import ODCConfig, ConfigException
    with pytest.raises(ConfigException) as e:
        cfg = ODCConfig(paths=path_to_ini_config, raw_dict=simple_dict, text=simple_config)
    assert "Can only supply one of" in str(e.value)
    with pytest.raises(ConfigException) as e:
        cfg = ODCConfig(raw_dict=simple_dict, text=simple_config)
    assert "Can only supply one of" in str(e.value)
    with pytest.raises(ConfigException) as e:
        cfg = ODCConfig(paths=path_to_ini_config, text=simple_config)
    assert "Can only supply one of" in str(e.value)
    with pytest.raises(ConfigException) as e:
        cfg = ODCConfig(paths=path_to_ini_config, raw_dict=simple_dict)
    assert "Can only supply one of" in str(e.value)


def test_get_environment(simple_config):
    from datacube.cfg import ODCConfig, ConfigException
    cfg = ODCConfig(text=simple_config)
    with pytest.raises(ConfigException) as e:
        cfg2 = ODCConfig.get_environment(config=cfg, raw_config=simple_config)
    assert "Cannot specify both" in str(e.value)
    env = ODCConfig.get_environment(config=cfg, env="default")
    assert env is cfg[None]


def test_raw_by_environment(simple_config, monkeypatch):
    from datacube.cfg import ODCConfig
    monkeypatch.setenv(
        'ODC_CONFIG',
        '{"default":{"alias": "foo"},"foo":{"index_driver":"postgis","db_url":"postgresql:///mydb"}}'
    )
    cfg = ODCConfig()
    assert cfg[None]._name == 'foo'


def test_default_environment(simple_config, monkeypatch):
    from datacube.cfg import ODCConfig
    cfg = ODCConfig(text=simple_config)
    assert cfg[None]._name == 'legacy'
    monkeypatch.setenv('ODC_ENVIRONMENT', 'exp2')
    assert cfg[None]._name == 'exp2'
    monkeypatch.setenv('ODC_ENVIRONMENT', '')
    monkeypatch.setenv('DATACUBE_ENVIRONMENT', 'postgis')
    assert cfg[None]._name == 'experimental'
    cfg = ODCConfig(raw_dict={
        'datacube': {"index_driver": "memory"}
    })
    monkeypatch.setenv('DATACUBE_ENVIRONMENT', '')
    assert cfg[None]._name == 'datacube'
    cfg = ODCConfig(raw_dict={
        'weirdname': {"index_driver": "memory"},
        'stupidenv': {"index_driver": "null"},
    })
    assert cfg[None]._name == "default"



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from time import monotonic
from typing import Mapping, Optional, Iterable
from uuid import UUID

from datacube.index.abstract import AbstractIndex, AbstractLineageResource, DSID, BatchStatus, dsid_to_uuid
from datacube.index.postgis._transaction import IndexResourceAddIn
from datacube.drivers.postgis._connections import PostGisDb
from datacube.model import LineageTree, LineageDirection, LineageRelation
from datacube.model.lineage import LineageRelations


class LineageResource(AbstractLineageResource, IndexResourceAddIn):
    def __init__(self, db: PostGisDb, index: AbstractIndex) -> None:
        """
        :type db: datacube.drivers.postgis._connections.PostgresDb
        :type metadata_type_resource: datacube.index._metadata_types.MetadataTypeResource
        """
        self._db = db
        super().__init__(index)

    def get_derived_tree(self, id_: DSID, max_depth: int = 0) -> LineageTree:
        return self.get_lineage_tree(id_, LineageDirection.DERIVED, max_depth)

    def get_source_tree(self, id_: DSID, max_depth: int = 0) -> LineageTree:
        return self.get_lineage_tree(id_, LineageDirection.SOURCES, max_depth)

    def get_lineage_tree(self, id_: DSID, direction: LineageDirection, max_depth: int):
        id_ = dsid_to_uuid(id_)
        with self._db_connection() as connection:
            # Extract lineage relations into a collection
            relations = connection.load_lineage_relations([id_],
                                                          direction,
                                                          max_depth)
            rels = LineageRelations(relations=relations)
            # Extract home information into the collection
            homes = connection.select_homes(rels.dataset_ids)
        for dsid, home in homes.items():
            rels.merge_new_home(dsid, home)
        # Extract tree from collection
        return rels.extract_tree(id_, direction)

    def add(self, tree: LineageTree, max_depth: int = 0, allow_updates: bool = False) -> None:
        # Convert to a relations collection
        relations = LineageRelations(tree=tree, max_depth=max_depth)
        # and merge into index.
        self.merge(relations, allow_updates=allow_updates)

    def merge(self, rels: LineageRelations, allow_updates: bool = False, validate_only: bool = False) -> None:
        if allow_updates and validate_only:
            raise ValueError("Cannot validate-only AND allow updates")
        with self._db_connection() as connection:
            # Get all current relations one step forwards and backwards from all dataset ids in the tree.
            db_relations = LineageRelations(
                relations=connection.get_all_relations(rels.dataset_ids),
                homes=connection.select_homes(rels.dataset_ids)
            )
            # Check for consistency:
            new_rels, update_rels, new_homes, update_homes = rels.relations_diff(
                existing_relations=db_relations,
                allow_updates=allow_updates
            )
            if validate_only:
                # If we get to here, data is safe to add
                return
            # Merge homes data
            if new_homes:
                homes_new: dict[str, list[UUID]] = {}
                for id_, home in new_homes.items():
                    if home in homes_new:
                        homes_new[home].append(id_)
                    else:
                        homes_new[home] = [id_]
                for home, ids in homes_new.items():
                    connection.insert_home(home, ids, allow_updates=False)
            if update_homes:
                homes_update: dict[str, list[UUID]] = {}
                for id_, home in update_homes.items():
                    if home in homes_update:
                        homes_update[home].append(id_)
                    else:
                        homes_update[home] = [id_]
                for home, ids in homes_update.items():
                    connection.insert_home(home, ids, allow_updates=allow_updates)
            # Merge Relations data
            rels_new = [
                LineageRelation(classifier=classifier, derived_id=ids.derived_id, source_id=ids.source_id)
                for ids, classifier in new_rels.items()
            ]
            rels_update = [
                LineageRelation(classifier=classifier, derived_id=ids.derived_id, source_id=ids.source_id)
                for ids, classifier in update_rels.items()
            ]
            connection.write_relations(rels_new, allow_updates=False)
            connection.write_relations(rels_update, allow_updates=True)

    def remove(self, id_: DSID, direction: LineageDirection, max_depth: int = 0) -> None:
        id_ = dsid_to_uuid(id_)
        with self._db_connection() as connection:
            # Convert tree to desired deoth to lineage relations collection
            relations = connection.load_lineage_relations([id_],
                                                          direction,
                                                          max_depth)
            rels = LineageRelations(relations=relations)
            if direction == LineageDirection.SOURCES:
                ids = list(rels.by_derived.keys())
            else:
                ids = list(rels.by_source.keys())
            # Delete individual relations.
            connection.remove_lineage_relations(ids, direction)

    def set_home(self, home: str, *args: DSID, allow_updates: bool = False) -> int:
        with self._db_connection() as connection:
            ids = (dsid_to_uuid(id_) for id_ in args)
            return connection.insert_home(home, ids, allow_updates)

    def clear_home(self, *args: DSID, home: Optional[str] = None) -> int:
        ids = [dsid_to_uuid(id_) for id_ in args]
        with self._db_connection() as connection:
            return connection.delete_home(ids)

    def get_homes(self, *args: DSID) -> Mapping[UUID, str]:
        ids = [dsid_to_uuid(id_) for id_ in args]
        with self._db_connection() as connection:
            return connection.select_homes(ids)

    def get_all_lineage(self, batch_size: int = 1000) -> Iterable[LineageRelation]:
        with self._db_connection(transaction=True) as connection:
            for row in connection.get_all_lineage(batch_size=batch_size):
                yield LineageRelation(
                    derived_id=row.derived_dataset_ref,
                    classifier=row.classifier,
                    source_id=row.source_dataset_ref
                )

    def _add_batch(self, batch: Iterable[LineageRelation]) -> BatchStatus:
        b_started = monotonic()
        with self._db_connection(transaction=True) as connection:
            b_added, b_skipped = connection.insert_lineage_bulk(
                [
                    {
                        "derived_dataset_ref": rel.derived_id,
                        "classifier": rel.classifier,
                        "source_dataset_ref": rel.source_id,
                    }
                    for rel in batch
                ]
            )
        return BatchStatus(b_added, b_skipped, monotonic() - b_started)



#!/usr/bin/env python
# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Query datasets.
"""

import csv
import datetime
import shutil
import sys
from functools import partial

import click
from sqlalchemy.dialects.postgresql import Range
from functools import singledispatch

from datacube.ui import click as ui
from datacube.ui.click import CLICK_SETTINGS

from datacube.utils.dates import tz_as_utc

PASS_INDEX = ui.pass_index('datacube-search')


def printable_values(d):
    return {k: printable(v) for k, v in d.items()}


def write_pretty(out_f, field_names, search_results, terminal_size=shutil.get_terminal_size()):
    """
    Output in a human-readable text format. Inspired by psql's expanded output.
    """
    terminal_width = terminal_size[0]
    record_num = 1

    field_header_width = max(len(name) for name in field_names)
    field_output_format = '{:<' + str(field_header_width) + '} | {}'

    for result in search_results:
        separator_line = '-[ {} ]'.format(record_num)
        separator_line += '-' * (terminal_width - len(separator_line) - 1)
        click.echo(separator_line, file=out_f)

        for name, value in sorted(result.items()):
            click.echo(
                field_output_format.format(name, printable(value)),
                file=out_f
            )

        record_num += 1


def write_csv(out_f, field_names, search_results):
    """
    Output as a CSV.
    """
    search_results = list(search_results)
    writer = csv.DictWriter(out_f, tuple(sorted(field_names)))
    writer.writeheader()
    writer.writerows(
        (
            printable_values(d) for d in
            search_results
        )
    )


OUTPUT_FORMATS = {
    'csv': write_csv,
    'pretty': write_pretty
}


@click.group(help="Search the Data Cube", context_settings=CLICK_SETTINGS)
@ui.global_cli_options
@click.option('-f',
              type=click.Choice(list(OUTPUT_FORMATS)),
              default='pretty', show_default=True,
              help='Output format')
@click.pass_context
def cli(ctx, f):
    ctx.obj['write_results'] = partial(OUTPUT_FORMATS[f], sys.stdout)


@cli.command()
@ui.parsed_search_expressions
@PASS_INDEX
@click.pass_context
def datasets(ctx, index, expressions):
    """
    Search available Datasets
    """
    ctx.obj['write_results'](
        sorted(index.products.get_field_names()),
        (tup._asdict() for tup in index.datasets.search_returning(**expressions))
    )


@cli.command('product-counts')
@click.argument('period', nargs=1)
@ui.parsed_search_expressions
@PASS_INDEX
def product_counts(index, period, expressions):
    """
    Count product Datasets available by period

    PERIOD: eg. 1 month, 6 months, 1 year
    """
    for product, series in index.datasets.count_by_product_through_time(period, **expressions):
        click.echo(product.name)
        for timerange, count in series:
            formatted_dt = tz_as_utc(timerange[0]).strftime("%Y-%m-%d")
            click.echo('    {}: {}'.format(formatted_dt, count))


@singledispatch
def printable(val):
    return val


@printable.register(type(None))
def printable_none(val):
    return ''


@printable.register(datetime.datetime)
def printable_dt(val):
    """
    :type val: datetime.datetime
    """
    return tz_as_utc(val).isoformat()


@printable.register(Range)
def printable_r(val):
    """
    :type val: sqlalchemy.dialects.postgresql.Range
    """
    if val.lower_inf:
        return printable(val.upper)
    if val.upper_inf:
        return printable(val.lower)

    return '{} to {}'.format(printable(val.lower), printable(val.upper))


if __name__ == '__main__':
    cli()
